"""
    This thermodynamic module contains a series of wrappers to handle the inputs and
    outputs of these functions.

    The ``calc`` module contains the thermodynamic calculations. Calculation of
    pressure, fugacity_coefficient, and maximum allowed density are handled by an Eos
    object so that these functions can be used with any EOS. None of the functions in
    this folder need to be handled directly, as a function factory is included in our
    ``__init__.py`` file. Add "from thermodynamics import thermo" and use
    `thermo("calc_type",Eos,input_dict)`` to get started.
"""

import numpy as np
import logging

from despasito.utils.parallelization import MultiprocessingJob
from . import calc
from despasito import fundamental_constants as constants
import despasito.utils.general_toolbox as gtb

logger = logging.getLogger(__name__)


def bubble_pressure(Eos, **sys_dict):
    r"""
    Calculates pressure and vapor compositions given liquid mole fractions and
    temperature.

    Parameters
    ----------
    Eos : obj
        An instance of the defined EOS class to be used in thermodynamic computations.
    xilist : list
        List of component mole fraction sets, where ``sum(xi)=1.0`` for each set. Each
        set of components corresponds to a temperature in Tlist, or if one set is
        given, this composition will be used for all temperatures.
    Tlist : list, Optional, default=298.15
        [K] Temperature of the system corresponding to composition in xilist. If one
        set is given, this temperature will be used for all compositions.
    Pguess : list/float, Optional
        [Pa] Guess the system pressure at the bubble point. If a list of length one
        is provided, that value is used for all temperature-composition sets given.
    Pmin : list, Optional
        [Pa] Set the upper bound for the minimum system pressure at the bubble point.
        If not defined, the default in
        :func:`~despasito.thermodynamics.calc.calc_bubble_pressure` is used. If a list
        of length one is provided, that value is used for all temperature-composition
        sets given.
    Pmax : list, Optional
        [Pa] Set the upper bound for the maximum system pressure at the bubble point.
        If not defined, the default in
        :func:`~despasito.thermodynamics.calc.calc_bubble_pressure` is used. If a list
        of length one is provided, that value is used for all temperature-composition
        sets given.
    MultiprocessingObject : obj, Optional
        Multiprocessing object,
        :class:`~despasito.utils.parallelization.MultiprocessingJob`
    kwargs : Optional
        Keyword arguments for
        :func:`~despasito.thermodynamics.calc.calc_saturation_properties` and
        :func:`~despasito.thermodynamics.calc.calc_bubble_pressure`

    Returns
    -------
    output_dict : dict
        Output dictionary containing given and calculated arrays

        - T: [K] Temperature array generated from given instructions
        - xi: Given array of liquid compositions
        - flagl: Phase flag for liquid given temperature and calculated pressure, a
        value of 0 is vapor, 1 is liquid, 2 mean a critical fluid, 3 means that
        neither is true, 4 means we should assume ideal gas
        - P: [Pa] Pressure values evaluated for given conditions
        - yi: Vapor mole fraction evaluated for given conditions
        - flagv: Phase flag for vapor given temperature and calculated pressure, a
        value of 0 is vapor, 1 is liquid, 2 mean a critical fluid, 3 means that
        neither is true, 4 means we should assume ideal gas
        - obj: List of objective values

    """

    if "xilist" not in sys_dict:
        raise ValueError("Mole fractions, xilist, are not specified")
    elif not gtb.isiterable(sys_dict["xilist"][0]):
        sys_dict["xilist"] = np.array([sys_dict["xilist"]], float)

    # Extract and check input data
    thermo_keys = ["Tlist", "xilist"]
    thermo_dict = gtb.check_length_dict(sys_dict, thermo_keys)
    npoints = len(thermo_dict[list(thermo_dict.keys())[0]])

    thermo_defaults = [constants.standard_temperature]
    thermo_dict.update(gtb.set_defaults(thermo_dict, "Tlist", thermo_defaults, lx=npoints))

    # Optional values
    optional_keys = ["Pguess", "Pmin", "Pmax"]
    opts = gtb.check_length_dict(sys_dict, optional_keys, lx=npoints)
    if opts:
        logger.info("Accepted user defined variables: {}".format(", ".join(list(opts.keys()))))

    # Delete processed keys
    for key in thermo_keys + optional_keys:
        if key in sys_dict:
            del sys_dict[key]

    if "MultiprocessingObject" in sys_dict:
        MultiprocessingObject = sys_dict["MultiprocessingObject"]
        flag_use_mp_object = True
        del sys_dict["MultiprocessingObject"]
    else:
        flag_use_mp_object = False

    opts.update(sys_dict)  # Add unprocessed options

    # Calculate P and yi
    per_job_var = ["Pmin", "Pmax", "Pguess"]
    inputs = []
    for i in range(npoints):
        opts_tmp = opts.copy()
        for key in per_job_var:
            if key in opts_tmp:
                opts_tmp[key] = opts_tmp[key][i]
        inputs.append((thermo_dict["Tlist"][i], thermo_dict["xilist"][i], Eos, opts_tmp))

    if flag_use_mp_object:
        P_list, yi_list, flagv_list, flagl_list, obj_list = MultiprocessingObject.pool_job(
            _bubble_pressure_wrapper, inputs
        )
    else:
        P_list, yi_list, flagv_list, flagl_list, obj_list = MultiprocessingJob.serial_job(
            _bubble_pressure_wrapper, inputs
        )

    logger.info("--- Calculation bubble_pressure Complete ---")

    return {
        "T": thermo_dict["Tlist"],
        "xi": thermo_dict["xilist"],
        "P": P_list,
        "yi": yi_list,
        "flagl": flagl_list,
        "flagv": flagv_list,
        "obj": obj_list,
    }


def _bubble_pressure_wrapper(args):
    r"""Wrapper for parallelized use of 'bubble_pressure' calculation type."""

    T, xi, Eos, opts = args
    logger.info("T (K), xi: {} {}, Let's Begin!".format(T, xi))

    try:
        if len(xi[xi != 0.0]) == 1:
            P, _, _ = calc.calc_saturation_properties(T, xi, Eos, **opts)
            yi, flagv, flagl, obj = xi, 0, 1, 0.0
        else:
            if "pressure_options" in opts and "method" in opts["pressure_options"]:
                opts["method"] = opts["pressure_options"]["method"]
                del opts["pressure_options"]["method"]
            P, yi, flagv, flagl, obj = calc.calc_bubble_pressure(xi, T, Eos, **opts)
    except Exception:
        logger.warning("T (K), xi: {} {}, calculation did not produce a valid result.".format(T, xi))
        logger.debug("Calculation Failed:", exc_info=True)
        P, yi, flagl, flagv, obj = [np.nan, np.nan * np.ones(len(xi)), 3, 3, np.nan]

    logger.info("P (Pa), yi: {} {}".format(P, yi))

    return P, yi, flagv, flagl, obj


def dew_pressure(Eos, **sys_dict):
    r"""
    Calculate pressure and liquid composition given vapor mole fractions, yi, and
    temperature.

    Parameters
    ----------
    Eos : obj
        An instance of the defined EOS class to be used in thermodynamic computations.
    yilist : list
        List of component mole fraction sets, where ``sum(yi)=1.0`` for each set.
        Each set of components corresponds to a temperature in Tlist, or if one set
        is given, this composition will be used for all temperatures.
    Tlist : list, Optional, default=298.15
        [K] Temperature of the system corresponding to composition in yilist. If one
        set is given, this pressure will be used for all compositions.
    Pguess : list/float, Optional
        [Pa] Guess the system pressure at the dew point. If a list of length one is
        provided, that value is used for all temperature-composition sets given.
    Pmin : list, Optional
        [Pa] Set the upper bound for the minimum system pressure at the dew point.
        If not defined, the default in
        :func:`~despasito.thermodynamics.calc.calc_dew_pressure` is used. If a list of
        length one is provided, that value is used for all temperature-composition
        sets given.
    Pmax : list, Optional
        [Pa] Set the upper bound for the maximum system pressure at the dew point.
        If not defined, the default in
        :func:`~despasito.thermodynamics.calc.calc_dew_pressure` is used. If a list
        of length one is provided, that value is used for all temperature-composition
        sets given.
    MultiprocessingObject : obj, Optional
        Multiprocessing object,
        :class:`~despasito.utils.parallelization.MultiprocessingJob`
    kwargs : Optional
        Keyword arguments for
        :func:`~despasito.thermodynamics.calc.calc_saturation_properties` and
        :func:`~despasito.thermodynamics.calc.calc_dew_pressure`

    Returns
    -------
    output_dict : dict
        Output dictionary containing given and calculated values

        - T: [K] Temperature array generated from given instructions
        - yi: Given array of vapor compositions
        - flagv: Phase flag for vapor given temperature and calculated pressure, a
        value of 0 is vapor, 1 is liquid, 2 mean a critical fluid, 3 means that
        neither is true, 4 means we should assume ideal gas
        - P: [Pa] Pressure values evaluated for given conditions
        - xi: Liquid mole fraction evaluated for given conditions
        - flagl: Phase flag for liquid given temperature and calculated pressure, a
        value of 0 is vapor, 1 is liquid, 2 mean a critical fluid, 3 means that
        neither is true, 4 means we should assume ideal gas
        - obj: List of objective values

    """

    if "yilist" not in sys_dict:
        raise ValueError("Mole fractions, yilist, are not specified")
    elif not gtb.isiterable(sys_dict["yilist"][0]):
        sys_dict["yilist"] = np.array([sys_dict["yilist"]], float)

    # Extract and check input data
    thermo_keys = ["Tlist", "yilist"]
    thermo_dict = gtb.check_length_dict(sys_dict, thermo_keys)
    npoints = len(thermo_dict[list(thermo_dict.keys())[0]])

    thermo_defaults = [constants.standard_temperature]
    thermo_dict.update(gtb.set_defaults(thermo_dict, "Tlist", thermo_defaults, lx=npoints))

    # Optional values
    optional_keys = ["Pguess", "Pmin", "Pmax"]
    opts = gtb.check_length_dict(sys_dict, optional_keys, lx=npoints)
    if opts:
        logger.info("Accepted user defined variables: {}".format(", ".join(list(opts.keys()))))

    # Delete processed keys
    for key in thermo_keys + optional_keys:
        if key in sys_dict:
            del sys_dict[key]

    if "MultiprocessingObject" in sys_dict:
        MultiprocessingObject = sys_dict["MultiprocessingObject"]
        del sys_dict["MultiprocessingObject"]
        flag_use_mp_object = True
    else:
        flag_use_mp_object = False

    opts.update(sys_dict)  # Add unprocessed options

    # Calculate P and yi
    per_job_var = ["Pguess", "Pmin", "Pmax"]
    inputs = []
    for i in range(npoints):
        opts_tmp = opts.copy()
        for key in per_job_var:
            if key in opts_tmp:
                opts_tmp[key] = opts_tmp[key][i]
        inputs.append((thermo_dict["Tlist"][i], thermo_dict["yilist"][i], Eos, opts_tmp))

    if flag_use_mp_object:
        P_list, xi_list, flagv_list, flagl_list, obj_list = MultiprocessingObject.pool_job(
            _dew_pressure_wrapper, inputs
        )
    else:
        P_list, xi_list, flagv_list, flagl_list, obj_list = MultiprocessingJob.serial_job(_dew_pressure_wrapper, inputs)

    logger.info("--- Calculation dew_pressure Complete ---")

    return {
        "T": thermo_dict["Tlist"],
        "xi": xi_list,
        "P": P_list,
        "yi": thermo_dict["yilist"],
        "flagl": flagl_list,
        "flagv": flagv_list,
        "obj": obj_list,
    }


def _dew_pressure_wrapper(args):
    r"""Wrapper for parallelized use of 'dew_pressure' calculation type."""

    T, yi, Eos, opts = args
    logger.info("T (K), yi: {} {}, Let's Begin!".format(T, yi))

    try:
        if len(yi[yi != 0.0]) == 1:
            P, _, _ = calc.calc_saturation_properties(T, yi, Eos, **opts)
            xi, flagv, flagl, obj = yi, 0, 1, 0.0
        else:
            if "pressure_options" in opts and "method" in opts["pressure_options"]:
                opts["method"] = opts["pressure_options"]["method"]
                del opts["pressure_options"]["method"]
            P, xi, flagl, flagv, obj = calc.calc_dew_pressure(yi, T, Eos, **opts)
    except Exception:
        logger.warning("T (K), yi: {} {}, calculation did not produce a valid result.".format(T, yi))
        logger.debug("Calculation Failed:", exc_info=True)
        P, xi, flagl, flagv, obj = [np.nan, np.nan * np.ones(len(yi)), 3, 3, np.nan]

    logger.info("P (Pa), xi: {} {}".format(P, xi))

    return P, xi, flagv, flagl, obj


def activity_coefficient(Eos, **sys_dict):
    r"""
    Calculate the activity coefficient with a variety of methods.

    Obtaining the activity coefficient requires the temperature, pressure, vapor
    composition, and liquid composition, but if one is missing, the necessary
    calculation computes it.

    Parameters
    ----------
    Eos : obj
        An instance of the defined EOS class to be used in thermodynamic computations.
    Tlist : list, Optional, default=298.15
        [K] Temperature of the system corresponding to composition in yilist. If one
        set is given, this pressure will be used for all compositions.
    xilist : list, Optional
        List of component mole fraction sets, where ``sum(yi)=1.0`` for each set.
        Each set of components corresponds to a temperature in Tlist, or if one set
        is given, this composition will be used for all temperatures. If this value
        is not provided, a dew point calculation is initiated if the pressure is also
        absent, and a flash calculation if yilist is absent.
    yilist : list, Optional
        List of component mole fraction sets, where ``sum(yi)=1.0`` for each set. Each
        set of components corresponds to a temperature in Tlist, or if one set is
        given, this composition will be used for all temperatures. If this value is
        not provided, a bubble point calculation is initiated if the pressure is also
        absent, and a flash calculation if xilist is absent.
    Plist : list, Optional
        [Pa] Pressure of the system corresponding to Tlist. If one value is given,
        this pressure will be used for all temperatures. If this value is not
        provided, a dew or bubble point calculation is initiated to find it.
    Pguess : list/float, Optional
        [Pa] If the pressure and either xilist or yilist is not provided, guess the
        system pressure at the dew/bubble point respectively. If a list of length one
        is provided, that value is used for all temperature-composition sets given.
    Pmin : list, Optional
        [Pa] If the pressure and either xilist or yilist is not provided, set a
        minimum bounds in pressure for the dew/bubble point respectively. If not
        defined, the default in :func:`~despasito.thermodynamics.calc.calc_dew_pressure`
        is used. If a list of length one is provided, that value is used for all
        temperature-composition sets given.
    Pmax : list, Optional
        [Pa] If the pressure and either xilist or yilist is not provided, set a maximum
        bounds in pressure for the dew/bubble point respectively. If not defined, the
        default in :func:`~despasito.thermodynamics.calc.calc_dew_pressure` is used.
        If a list of length one is provided, that value is used for all
        temperature-composition sets given.
    MultiprocessingObject : obj, Optional
        Multiprocessing object,
        :class:`~despasito.utils.parallelization.MultiprocessingJob`
    kwargs : Optional
        Keyword arguments for
        :func:`~despasito.thermodynamics.calc.calc_saturation_properties` and either
        :func:`~despasito.thermodynamics.calc.calc_dew_pressure`,
        :func:`~despasito.thermodynamics.calc.calc_bubble_pressure`, or
        :func:`~despasito.thermodynamics.calc.calc_flash`, depending on which two of
        xi, yi, or P is missing.

    Returns
    -------
    output_dict : dict
        Output dictionary containing given and calculated values

        - T: [K] Temperature array generated from given instructions
        - P: [Pa] Given or calculated array of pressures
        - xi: Given or calculated array of liquid compositions
        - yi: Given or calculated array of vapor compositions
        - flagv: Phase flag for vapor given temperature and calculated pressure, a
        value of 0 is vapor, 1 is liquid, 2 mean a critical fluid, 3 means that
        neither is true, 4 means we should assume ideal gas
        - flagl: Phase flag for liquid given temperature and calculated pressure, a
        value of 0 is vapor, 1 is liquid, 2 mean a critical fluid, 3 means that
        neither is true, 4 means we should assume ideal gas
        - obj: List of objective values if a calculation was needed
        - Psat: [Pa] List of calculated saturation pressures
        - gamma: List of calculated activity coefficients

    """

    if "xilist" in sys_dict and not gtb.isiterable(sys_dict["xilist"][0]):
        sys_dict["xilist"] = np.array([sys_dict["xilist"]], float)
    if "yilist" in sys_dict and not gtb.isiterable(sys_dict["yilist"][0]):
        sys_dict["yilist"] = np.array([sys_dict["yilist"]], float)

    # Extract and check input data
    thermo_keys = ["Tlist", "yilist", "xilist", "Plist"]
    thermo_dict = gtb.check_length_dict(sys_dict, thermo_keys)
    npoints = len(thermo_dict[list(thermo_dict.keys())[0]])

    # Determine Mode
    if "xilist" not in thermo_dict:
        if "yilist" not in thermo_dict:
            if "Plist" in thermo_dict:
                mode = "flash"
            else:
                mode = None
        else:
            if "Plist" in thermo_dict:
                thermo_dict["Pguess"] = thermo_dict["Plist"]
                logger.info("Using given pressure as initial guess")
            mode = "dew_point"
    elif "yilist" not in thermo_dict:
        if "Plist" in thermo_dict:
            thermo_dict["Pguess"] = thermo_dict["Plist"]
            logger.info("Using given pressure as initial guess")
        mode = "bubble_point"
    else:
        mode = "standard"

    if mode is None:
        raise ValueError("Two of the following system properties must be provided: Tlist, Plist," " xilist, or yilist")
    else:
        logger.info("Activity coefficient being calculated in {} mode.".format(mode))

    # Set Defaults
    thermo_defaults = [
        constants.standard_temperature,
        np.nan * np.ones(Eos.number_of_components),
        np.nan * np.ones(Eos.number_of_components),
        np.nan,
    ]
    thermo_dict.update(gtb.set_defaults(thermo_dict, thermo_keys, thermo_defaults, lx=npoints))

    # Optional values
    optional_keys = ["Pguess", "Pmin", "Pmax"]
    opts = gtb.check_length_dict(sys_dict, optional_keys, lx=npoints)
    if opts:
        logger.info("Accepted user defined variables: {}".format(", ".join(list(opts.keys()))))

    # Delete processed keys
    for key in thermo_keys + optional_keys:
        if key in sys_dict:
            del sys_dict[key]

    if "MultiprocessingObject" in sys_dict:
        MultiprocessingObject = sys_dict["MultiprocessingObject"]
        del sys_dict["MultiprocessingObject"]
        flag_use_mp_object = True
    else:
        flag_use_mp_object = False

    opts.update(sys_dict)  # Add unprocessed options

    # Calculate P and yi
    per_job_var = ["Pguess", "Pmin", "Pmax"]
    inputs = []
    for i in range(npoints):
        opts_tmp = opts.copy()
        for key in per_job_var:
            if key in opts_tmp:
                opts_tmp[key] = opts_tmp[key][i]
        inputs.append(
            (
                thermo_dict["Tlist"][i],
                thermo_dict["Plist"][i],
                thermo_dict["xilist"][i],
                thermo_dict["yilist"][i],
                Eos,
                opts_tmp,
                mode,
            )
        )

    if flag_use_mp_object:
        tmp = MultiprocessingObject.pool_job(_activity_coefficient_wrapper, inputs)
    else:
        tmp = MultiprocessingJob.serial_job(_activity_coefficient_wrapper, inputs)
    (
        T_list,
        P_list,
        xi_list,
        yi_list,
        flagv_list,
        flagl_list,
        obj_list,
        Psat_list,
        gamma_list,
    ) = tmp

    logger.info("--- Calculation dew_pressure Complete ---")

    return {
        "T": T_list,
        "xi": xi_list,
        "P": P_list,
        "yi": yi_list,
        "flagl": flagl_list,
        "flagv": flagv_list,
        "obj": obj_list,
        "Psat": Psat_list,
        "gamma": gamma_list,
    }


def _activity_coefficient_wrapper(args):
    r"""Wrapper for parallelized use of 'dew_pressure' calculation type."""

    T, P, xi, yi, Eos, opts, mode = args
    logger.info("T (K), yi: {} {}, Let's Begin!".format(T, yi))

    try:
        if "pressure_options" in opts and "method" in opts["pressure_options"]:
            opts["method"] = opts["pressure_options"]["method"]
            del opts["pressure_options"]["method"]

        if mode == "dew_point":
            P, xi, flagl, flagv, obj = calc.calc_dew_pressure(yi, T, Eos, **opts)
        elif mode == "bubble_point":
            P, yi, flagv, flagl, obj = calc.calc_bubble_pressure(xi, T, Eos, **opts)
        elif mode == "flash":
            xi, flagl, yi, flagv, obj = calc.calc_flash(P, T, Eos, **opts)
        elif mode == "standard":
            flagl, flagv, obj = [np.nan, np.nan, np.nan]

        gamma, Psat = calc.activity_coefficient(P, T, xi, yi, Eos, **opts)

    except Exception:
        logger.warning(
            "In calculating the activity coefficient, {} ".format(mode)
            + "calculation with T, P, xi, and yi inputs: "
            + "{}, {}, {}, {} did not produce a valid result.".format(T, P, xi, yi)
        )
        logger.debug("Calculation Failed:", exc_info=True)

        if mode == "dew_point":
            P, xi, flagl, flagv, obj, Psat, gamma = [
                np.nan,
                np.nan * np.ones(len(yi)),
                3,
                3,
                np.nan,
                np.nan,
                np.nan,
            ]
        elif mode == "bubble_point":
            P, yi, flagl, flagv, obj, Psat, gamma = [
                np.nan,
                np.nan * np.ones(len(xi)),
                3,
                3,
                np.nan,
                np.nan,
                np.nan,
            ]
        elif mode == "flash":
            xi, yi, flagl, flagv, obj, Psat, gamma = [
                np.nan * np.ones(Eos.number_of_components),
                np.nan * np.ones(Eos.number_of_components),
                3,
                3,
                np.nan,
                np.nan,
                np.nan,
            ]
        elif mode == "standard":
            flagl, flagv, obj, Psat, gamma = [np.nan, np.nan, np.nan, np.nan, np.nan]

    logger.info("Psat (Pa), activity coefficient: {} {}".format(Psat, gamma))

    return T, P, xi, yi, flagv, flagl, obj, Psat, gamma


def flash(Eos, **sys_dict):
    r"""
    Flash calculation of vapor and liquid mole fractions. Only binary systems are
    currently supported

    Input and system information are assessed first. An output file is generated with
    T, yi, and corresponding P and xi.

    Parameters
    ----------
    Eos : obj
        An instance of the defined EOS class to be used in thermodynamic computations.
    Tlist : list, Optional, default=298.15
        [K] Temperature of the system corresponding Plist. If one value is given, this
        temperature will be used for all temperatures.
    Plist : list, Optional, default=101325.0
        [Pa] Pressure of the system corresponding to Tlist. If one value is given,
        this pressure will be used for all temperatures.
    MultiprocessingObject : obj, Optional
        Multiprocessing object,
        :class:`~despasito.utils.parallelization.MultiprocessingJob`
    kwargs : Optional
        Keyword arguments for :func:`~despasito.thermodynamics.calc.calc_flash`

    Returns
    -------
    output_dict : dict
        Output dictionary containing given and calculated values

        - T: [K] Temperature array generated from given instructions
        - P: [Pa] Pressure array generated from given instructions
        - xi: Given array of liquid compositions
        - flagl: Phase flag for liquid given temperature and calculated pressure, a
        value of 0 is vapor, 1 is liquid, 2 mean a critical fluid, 3 means that
        neither is true, 4 means we should assume ideal gas
        - yi: Vapor mole fraction evaluated for given conditions
        - flagv: Phase flag for vapor given temperature and calculated pressure,
        a value of 0 is vapor, 1 is liquid, 2 mean a critical fluid, 3 means that
        neither is true, 4 means we should assume ideal gas
        - obj: List of objective values

    """

    # Extract and check input data
    thermo_keys = ["Tlist", "Plist"]
    thermo_dict = gtb.check_length_dict(sys_dict, thermo_keys)
    npoints = len(thermo_dict[list(thermo_dict.keys())[0]])

    thermo_defaults = [constants.standard_temperature, constants.standard_pressure]
    thermo_dict.update(gtb.set_defaults(thermo_dict, thermo_keys, thermo_defaults, lx=npoints))

    # Delete processed keys
    for key in thermo_keys:
        if key in sys_dict:
            del sys_dict[key]

    if "MultiprocessingObject" in sys_dict:
        MultiprocessingObject = sys_dict["MultiprocessingObject"]
        del sys_dict["MultiprocessingObject"]
        flag_use_mp_object = True
    else:
        flag_use_mp_object = False

    opts = sys_dict.copy()

    if Eos.number_of_components != 2:
        raise ValueError(
            "Only binary systems are currently supported for flash calculations, "
            "{} were given.".format(Eos.number_of_components)
        )

    inputs = [(thermo_dict["Tlist"][i], thermo_dict["Plist"][i], Eos, opts) for i in range(npoints)]

    if flag_use_mp_object:
        xi_list, yi_list, flagv_list, flagl_list, obj_list = MultiprocessingObject.pool_job(_flash_wrapper, inputs)
    else:
        xi_list, yi_list, flagv_list, flagl_list, obj_list = MultiprocessingJob.serial_job(_flash_wrapper, inputs)

    logger.info("--- Calculation flash Complete ---")

    return {
        "T": thermo_dict["Tlist"],
        "xi": xi_list,
        "P": thermo_dict["Plist"],
        "yi": yi_list,
        "flagl": flagl_list,
        "flagv": flagv_list,
        "obj": obj_list,
    }


def _flash_wrapper(args):
    r"""Wrapper for parallelized use of 'flash' calculation type."""

    T, P, Eos, opts = args

    try:
        xi, flagl, yi, flagv, obj = calc.calc_flash(P, T, Eos, **opts)
    except Exception:
        logger.warning("T (K), P (Pa): {} {}, calculation".format(T, P) + " did not produce a valid result.")
        logger.debug("Calculation Failed:", exc_info=True)
        xi, yi, flagl, flagv, obj = [
            np.nan * np.ones(Eos.number_of_components),
            np.nan * np.ones(Eos.number_of_components),
            3,
            3,
            np.nan,
        ]

    logger.info("xi: {}, yi: {}".format(xi, yi))

    return xi, yi, flagv, flagl, obj


def saturation_properties(Eos, **sys_dict):
    r"""
    Computes the saturated pressure, liquid, and gas density a one component phase
    at a temperature.

    Input and system information are assessed first.  An output file is generated
    with T, :math:`P^{sat}`, :math:`\rho^{sat}_{l}`, and :math:`\rho^{sat}_{v}`.

    Parameters
    ----------
    Eos : obj
        An instance of the defined EOS class to be used in thermodynamic computations.
    Tlist : list, Optional, default=298.15
        [K] Temperature of the system corresponding Plist. If one value is given, this
        temperature will be used for all temperatures.
    MultiprocessingObject : obj, Optional
        Multiprocessing object,
        :class:`~despasito.utils.parallelization.MultiprocessingJob`
    kwargs : Optional
        Keyword arguments for
        :func:`~despasito.thermodynamics.calc.calc_saturation_properties`

    Returns
    -------
    output_dict : dict
        Output dictionary containing given and calculated values

        - T: [K] Temperature array generated from given instructions
        - Psat: [Pa] Saturation pressure
        - rhol: [mol/:math:`m^3`] Liquid saturation density
        - rholv: [mol/:math:`m^3`] Vapor saturation density

    """

    # Extract and check input data

    if "xilist" in sys_dict and not gtb.isiterable(sys_dict["xilist"][0]):
        sys_dict["xilist"] = np.array([sys_dict["xilist"]], float)

    thermo_keys = ["Tlist", "xilist"]
    thermo_dict = gtb.check_length_dict(sys_dict, thermo_keys)
    if "Tlist" not in thermo_dict:
        thermo_dict["Tlist"] = np.array([constants.standard_temperature])
        logger.info("Assuming standard temperature, {}".format(constants.standard_temperature))
    npoints = len(thermo_dict[list(thermo_dict.keys())[0]])

    thermo_defaults = [
        constants.standard_temperature,
        np.array([[1.0] for x in range(npoints)]),
    ]
    thermo_dict.update(gtb.set_defaults(thermo_dict, thermo_keys, thermo_defaults, lx=npoints))

    # Delete processed keys
    for key in thermo_keys:
        if key in sys_dict:
            del sys_dict[key]

    if "MultiprocessingObject" in sys_dict:
        MultiprocessingObject = sys_dict["MultiprocessingObject"]
        del sys_dict["MultiprocessingObject"]
        flag_use_mp_object = True
    else:
        flag_use_mp_object = False

    opts = sys_dict.copy()

    inputs = [(thermo_dict["Tlist"][i], thermo_dict["xilist"][i], Eos, opts) for i in range(npoints)]
    if flag_use_mp_object:
        Psat, rholsat, rhovsat = MultiprocessingObject.pool_job(_saturation_properties_wrapper, inputs)
    else:
        Psat, rholsat, rhovsat = MultiprocessingJob.serial_job(_saturation_properties_wrapper, inputs)

    logger.info("--- Calculation saturation_properties Complete ---")

    return {"T": thermo_dict["Tlist"], "Psat": Psat, "rhol": rholsat, "rhov": rhovsat}


def _saturation_properties_wrapper(args):
    r"""Wrapper for parallelized use of 'saturation_properties' calculation type."""

    T, xi, Eos, opts = args

    logger.info("T (K), xi: {} {}, Let's Begin!".format(T, xi))

    try:
        Psat, rholsat, rhovsat = calc.calc_saturation_properties(T, xi, Eos, **opts)
        if np.isnan(Psat):
            logger.warning("T (K), xi: {} {}, calculation did not produce a valid result.".format(T, xi))
            logger.debug("Calculation Failed:", exc_info=True)
            Psat, rholsat, rhovsat = [np.nan, np.nan, np.nan]
        else:
            logger.info("Psat {} Pa, rhol {}, rhov {}".format(Psat, rholsat, rhovsat))
    except Exception:
        logger.warning("T (K), xi: {} {}, calculation did not produce a valid result.".format(T, xi))
        logger.debug("Calculation Failed:", exc_info=True)
        Psat, rholsat, rhovsat = [np.nan, np.nan, np.nan]

    return Psat, rholsat, rhovsat


def liquid_properties(Eos, **sys_dict):
    r"""
    Computes the liquid density and fugacity coefficient given a temperature, pressure,
    and liquid mole fractions.

    Input and system information are assessed first. An output file is generated with P,
    T, xi, :math:`\rho_{l}`, and :math:`\phi_{l}`.

    Parameters
    ----------
    Eos : obj
        An instance of the defined EOS class to be used in thermodynamic computations.
    xilist : list
        List of component mole fraction sets, where ``sum(xi)=1.0`` for each set. Each
        set of components corresponds to a temperature in Tlist, or if one set is
        given, this composition will be used for all temperatures.
    Tlist : list, Optional default=298.15
        [K] Temperature of the system corresponding Plist. If one value is given, this
        temperature will be used for all temperatures.
    Plist : list, Optional, default=101325.0
        [Pa] Pressure of the system corresponding to Tlist. If one value is given,
        this pressure will be used for all temperatures.
    MultiprocessingObject : obj, Optional
        Multiprocessing object,
        :class:`~despasito.utils.parallelization.MultiprocessingJob`
    kwargs : Optional
        Keyword arguments for
        :func:`~despasito.thermodynamics.calc.calc_liquid_fugacity_coefficient`

    Returns
    -------
    output_dict : dict
        Output dictionary containing given and calculated values

        - T: [K] Temperature array generated from given instructions
        - P: [Pa] Pressure array generated from given instructions
        - xi: Composition generated from given instructions
        - rhol: [mol/:math:`m^3`] Evaluated liquid density
        - phil: Evaluated liquid fugacity coefficient
        - flagl: Phase flag for liquid given temperature and calculated pressure, a
        value of 0 is vapor, 1 is liquid, 2 mean a critical fluid, 3 means that
        neither is true, 4 means we should assume ideal gas

    """

    # Extract and check input data

    if "xilist" in sys_dict and not gtb.isiterable(sys_dict["xilist"][0]):
        sys_dict["xilist"] = np.array([sys_dict["xilist"]], float)

    thermo_keys = ["Tlist", "xilist", "Plist"]
    thermo_dict = gtb.check_length_dict(sys_dict, thermo_keys)
    npoints = len(thermo_dict[list(thermo_dict.keys())[0]])

    if "xilist" not in thermo_dict and Eos.number_of_components > 1:
        raise ValueError("Ambiguous mixture composition. Define xilist")

    thermo_defaults = [
        constants.standard_temperature,
        np.array([[1.0] for x in range(npoints)]),
        constants.standard_pressure,
    ]
    thermo_dict.update(gtb.set_defaults(thermo_dict, thermo_keys, thermo_defaults, lx=npoints))

    # Delete processed keys
    for key in thermo_keys:
        if key in sys_dict:
            del sys_dict[key]

    if "MultiprocessingObject" in sys_dict:
        MultiprocessingObject = sys_dict["MultiprocessingObject"]
        del sys_dict["MultiprocessingObject"]
        flag_use_mp_object = True
    else:
        flag_use_mp_object = False

    opts = sys_dict.copy()

    inputs = [
        (
            thermo_dict["Plist"][i],
            thermo_dict["Tlist"][i],
            thermo_dict["xilist"][i],
            Eos,
            opts,
        )
        for i in range(npoints)
    ]
    if flag_use_mp_object:
        rhol, phil, flagl = MultiprocessingObject.pool_job(_liquid_properties_wrapper, inputs)
    else:
        rhol, phil, flagl = MultiprocessingJob.serial_job(_liquid_properties_wrapper, inputs)

    logger.info("--- Calculation liquid_properties Complete ---")

    return {
        "P": thermo_dict["Plist"],
        "T": thermo_dict["Tlist"],
        "xi": thermo_dict["xilist"],
        "rhol": rhol,
        "phil": phil,
        "flagl": flagl,
    }


def _liquid_properties_wrapper(args):
    r"""Wrapper for parallelized use of 'liquid_properties' calculation type."""

    P, T, xi, Eos, opts = args

    logger.info("T (K), P (Pa), xi: {} {} {}, Let's Begin!".format(T, P, xi))

    try:
        phil, rhol, flagl = calc.calc_liquid_fugacity_coefficient(P, T, xi, Eos, **opts)
        logger.info("P {} Pa, T {} K, xi {}, rhol {}, phil {}, flagl {}".format(P, T, xi, rhol, phil, flagl))
    except Exception:
        logger.warning("Failed to calculate rhol at {} K and {} Pa".format(T, P))
        rhol, flagl = np.nan, 3
        phil = np.nan * np.ones(Eos.number_of_components)
        logger.debug("Calculation Failed:", exc_info=True)

    return rhol, phil, flagl


def vapor_properties(Eos, **sys_dict):
    r"""
    Computes the vapor density and fugacity coefficient given a temperature, pressure,
    and vapor mole fractions.

    Input and system information are assessed first. An output file is generated with
    P, T, yi, :math:`\rho_{v}`, and :math:`\phi_{v}`.

    Parameters
    ----------
    Eos : obj
        An instance of the defined EOS class to be used in thermodynamic computations.
    yilist: : list
        List of component mole fraction sets, where ``sum(yi)=1.0`` for each set. Each
        set of components corresponds to a temperature in Tlist, or if one set is
        given, this composition will be used for all temperatures.
    Tlist : list, Optional, default=298.15
        [K] Temperature of the system corresponding Plist. If one value is given, this
        temperature will be used for all temperatures.
    Plist : list, Optional, default=101325.0
        [Pa] Pressure of the system corresponding to Tlist. If one value is given,
        this pressure will be used for all temperatures.
    MultiprocessingObject : obj, Optional
        Multiprocessing object,
        :class:`~despasito.utils.parallelization.MultiprocessingJob`
    kwargs : Optional
        Keyword arguments for
        :func:`~despasito.thermodynamics.calc.calc_vapor_fugacity_coefficient`

    Returns
    -------
    output_dict : dict
        Output dictionary containing given and calculated values

        - T: [K] Temperature array generated from given instructions
        - P: [Pa] Pressure array generated from given instructions
        - yi: Composition generated from given instructions
        - rhov: [mol/:math:`m^3`] Evaluated vapor density
        - phiv: Evaluated vapor fugacity coefficient
        - flagv: Phase flag for vapor given temperature and calculated pressure, a
        value of 0 is vapor, 1 is liquid, 2 mean a critical fluid, 3 means that
        neither is true, 4 means we should assume ideal gas

    """

    # Extract and check input data

    if "yilist" in sys_dict and not gtb.isiterable(sys_dict["yilist"][0]):
        sys_dict["yilist"] = np.array([sys_dict["yilist"]], float)

    thermo_keys = ["Tlist", "yilist", "Plist"]
    thermo_dict = gtb.check_length_dict(sys_dict, thermo_keys)
    npoints = len(thermo_dict[list(thermo_dict.keys())[0]])

    if "yilist" not in thermo_dict and Eos.number_of_components > 1:
        raise ValueError("Ambiguous mixture composition. Define yilist")

    thermo_defaults = [
        constants.standard_temperature,
        np.array([[1.0] for x in range(npoints)]),
        constants.standard_pressure,
    ]
    thermo_dict.update(gtb.set_defaults(thermo_dict, thermo_keys, thermo_defaults, lx=npoints))

    # Delete processed keys
    for key in thermo_keys:
        if key in sys_dict:
            del sys_dict[key]

    if "MultiprocessingObject" in sys_dict:
        MultiprocessingObject = sys_dict["MultiprocessingObject"]
        del sys_dict["MultiprocessingObject"]
        flag_use_mp_object = True
    else:
        flag_use_mp_object = False

    opts = sys_dict.copy()

    inputs = [
        (
            thermo_dict["Plist"][i],
            thermo_dict["Tlist"][i],
            thermo_dict["yilist"][i],
            Eos,
            opts,
        )
        for i in range(npoints)
    ]
    if flag_use_mp_object:
        rhov, phiv, flagv = MultiprocessingObject.pool_job(_vapor_properties_wrapper, inputs)
    else:
        rhov, phiv, flagv = MultiprocessingJob.serial_job(_vapor_properties_wrapper, inputs)

    logger.info("--- Calculation vapor_properties Complete ---")

    return {
        "P": thermo_dict["Plist"],
        "T": thermo_dict["Tlist"],
        "yi": thermo_dict["yilist"],
        "rhov": rhov,
        "phiv": phiv,
        "flagv": flagv,
    }


def _vapor_properties_wrapper(args):
    r"""Wrapper for parallelized use of 'vapor_properties' calculation type."""

    P, T, yi, Eos, opts = args

    logger.info("T (K), P (Pa), yi: {} {} {}, Let's Begin!".format(T, P, yi))

    try:
        phiv, rhov, flagv = calc.calc_vapor_fugacity_coefficient(P, T, yi, Eos, **opts)
        logger.info("P {} Pa, T {} K, yi {}, rhov {}, phiv {}, flagv {}".format(P, T, yi, rhov, phiv, flagv))
    except Exception:
        logger.warning("Failed to calculate rhov at {} K and {} Pa".format(T, P))
        rhov, flagv = np.nan, 3
        phiv = np.nan * np.ones(Eos.number_of_components)
        logger.debug("Calculation Failed:", exc_info=True)

    return rhov, phiv, flagv


def solubility_parameter(Eos, **sys_dict):
    r"""
    Calculate the Hildebrand solubility parameter based on temperature and composition.

    This function is based on the method used in Zeng, Z., Y. Xi, and Y. Li
    *Calculation of Solubility Parameter Using Perturbed-Chain SAFT and*
    *Cubic-Plus-Association Equations of State* Ind. Eng. Chem. Res. 2008. 47,
    9663-9669.

    Input and system information are assessed first. An output file is generated
    with T, xi, :math:`\rho_{l}`, and :math:`\delta`.

    Parameters
    ----------
    Eos : obj
        An instance of the defined EOS class to be used in thermodynamic computations.
    Tlist : list, Optional, default=298.15
        [K] Temperature of the system corresponding Plist. If one value is given, this
        temperature will be used for all temperatures.
    xilist : list, Optional, default=[1.0]
        Default assumes all of one component. List of component mole fraction sets,
        where ``sum(xi)=1.0`` for each set. Each set of components corresponds to a
        temperature in Tlist, or if one set is given, this composition will be used
        for all temperatures.
    Plist : list, Optional, default=101325.0
        [Pa] Pressure of the system corresponding to Tlist. If one value is given,
        this pressure will be used for all temperatures.
    MultiprocessingObject : obj, Optional
        Multiprocessing object,
        :class:`~despasito.utils.parallelization.MultiprocessingJob`
    kwargs : Optional
        Keyword arguments for
        :func:`~despasito.thermodynamics.calc.calc_liquid_density` and
        :func:`~despasito.thermodynamics.calc.calc_hildebrand_solubility`

    Returns
    -------
    output_dict : dict
        Output dictionary containing given and calculated values

        - T: [K] Temperature array generated from given instructions
        - P: [Pa] Pressure array generated from given instructions
        - xi: Composition generated from given instructions
        - rhol: [mol/:math:`m^3`] Evaluated liquid density
        - delta: Hidebrand solubility parameter given system conditions

    """

    if "xilist" in sys_dict and not gtb.isiterable(sys_dict["xilist"][0]):
        sys_dict["xilist"] = np.array([sys_dict["xilist"]], float)

    # Extract and check input data
    thermo_keys = ["Tlist", "xilist", "Plist"]
    if any([x in sys_dict for x in thermo_keys]):
        thermo_dict = gtb.check_length_dict(sys_dict, thermo_keys)
        npoints = len(thermo_dict[list(thermo_dict.keys())[0]])
    else:
        npoints = 1

    if "xilist" not in thermo_dict and Eos.number_of_components > 1:
        raise ValueError("Ambiguous mixture composition. Define xilist")

    thermo_defaults = [
        constants.standard_temperature,
        np.array([[1.0] for x in range(npoints)]),
        constants.standard_pressure,
    ]
    thermo_dict.update(gtb.set_defaults(thermo_dict, thermo_keys, thermo_defaults, lx=npoints))

    # Delete processed keys
    for key in thermo_keys:
        if key in sys_dict:
            del sys_dict[key]

    if "MultiprocessingObject" in sys_dict:
        MultiprocessingObject = sys_dict["MultiprocessingObject"]
        del sys_dict["MultiprocessingObject"]
        flag_use_mp_object = True
    else:
        flag_use_mp_object = False

    opts = sys_dict.copy()

    # Calculate solubility parameter
    inputs = [
        (
            thermo_dict["Plist"][i],
            thermo_dict["Tlist"][i],
            thermo_dict["xilist"][i],
            Eos,
            opts,
        )
        for i in range(npoints)
    ]
    if flag_use_mp_object:
        rhol, flagl, delta = MultiprocessingObject.pool_job(_solubility_parameter_wrapper, inputs)
    else:
        rhol, flagl, delta = MultiprocessingJob.serial_job(_solubility_parameter_wrapper, inputs)

    logger.info("--- Calculation solubility_parameter Complete ---")

    return {
        "P": thermo_dict["Plist"],
        "T": thermo_dict["Tlist"],
        "xi": thermo_dict["xilist"],
        "rhol": rhol,
        "delta": delta,
    }


def _solubility_parameter_wrapper(args):
    r"""Wrapper for parallelized use of 'solubility_parameter' calculation type."""

    P, T, xi, Eos, opts = args

    logger.info("T (K), P (Pa), xi: {} {} {}, Let's Begin!".format(T, P, xi))

    try:
        rhol, flagl = calc.calc_liquid_density(P, T, xi, Eos, **opts)
        delta = calc.hildebrand_solubility(rhol, xi, T, Eos, **opts)
        logger.info("P {} Pa, T {} K, xi {}, rhol {}, flagl {}, delta {}".format(P, T, xi, rhol, flagl, delta))
    except Exception:
        logger.warning("Failed to calculate rhov at {} K and {} Pa".format(T, P))
        rhol, flagl, delta = np.nan, 3, np.nan
        logger.debug("Calculation Failed:", exc_info=True)

    return rhol, flagl, delta


def verify_eos(Eos, **sys_dict):
    r"""
    The following consistency checks are performed to ensure the calculated fugacity
    coefficients are thermodynamically consistent.

    - 1. :math:`d(ln φ)/dP = (Z-1)/P`
    - 2. :math:`\sum{x_i d(ln φ)/d n_1 )} = 0`

    Parameters
    ----------
    Eos : obj
        An instance of the defined EOS class to be used in thermodynamic computations.
    Tlist : list, Optional, default=298.15
        [K] Temperature of the system corresponding Plist. If one value is given, this
        temperature will be used for all temperatures.
    xilist : list, Optional
        Default array of 11 values from x1=0 to x1=1 for binary array. List of
        component mole fraction sets, where ``sum(xi)=1.0`` for each set. Each set
        of components corresponds to a temperature in Tlist, or if one set is given,
        this composition will be used for all temperatures.
    Plist : list, Optional, default=101325.0
        [Pa] Pressure of the system corresponding to Tlist. If one value is given,
        this pressure will be used for all temperatures.
    MultiprocessingObject : obj, Optional
        Multiprocessing object,
        :class:`~despasito.utils.parallelization.MultiprocessingJob`
    kwargs : Optional
        Keyword arguments for
        :func:`~despasito.thermodynamics.calc.calc_vapor_density`,
        :func:`~despasito.thermodynamics.calc.calc_liquid_density`,
        :func:`~despasito.thermodynamics.calc.calc_fugacity_test_1`, and
        :func:`~despasito.thermodynamics.calc.calc_fugacity_test_2`,

    Returns
    -------
    output_dict : dict
        Output dictionary containing given and calculated values

        - T: [K] Temperature array generated from given instructions
        - P: [Pa] Pressure array generated from given instructions
        - xi: Composition generated from given instructions
        - residual_v1: Residual of vapor from method one
        - residual_v2: Residual of vapor from method two
        - flagv: Phase flag for vapor given temperature and calculated pressure,
        a value of 0 is vapor, 1 is liquid, 2 mean a critical fluid, 3 means that
        neither is true, 4 means we should assume ideal gas
        - log_phivi: Log of the partial vapor fugacity coefficient for each component
        - residual_l1: Residual of liquid from method one
        - residual_l2: Residual of liquid from method two
        - flagl: Phase flag for liquid given temperature and calculated pressure, a
        value of 0 is vapor, 1 is liquid, 2 mean a critical fluid, 3 means that
        neither is true, 4 means we should assume ideal gas
        - log_phili: Log of the partial liquid fugacity coefficient for each component

    """

    # Extract and check input data
    if "xilist" not in sys_dict:
        if Eos.number_of_components == 2:
            tmp = np.linspace(0.01, 0.99, 11)
            sys_dict["xilist"] = np.array([[x, 1.0 - x] for x in tmp])
            logger.info("Use array of mole fractions")
        else:
            raise ValueError(
                "Must have at least 2 components. With more that 2 components, the"
                " mole fractions need to be specified"
            )
    elif "xilist" in sys_dict and not gtb.isiterable(sys_dict["xilist"][0]):
        sys_dict["xilist"] = np.array([sys_dict["xilist"]], float)

    thermo_keys = ["Tlist", "xilist", "Plist"]
    thermo_dict = gtb.check_length_dict(sys_dict, thermo_keys)
    npoints = len(thermo_dict[list(thermo_dict.keys())[0]])

    thermo_defaults = [constants.standard_temperature, constants.standard_pressure]
    thermo_dict.update(gtb.set_defaults(thermo_dict, ["Tlist", "Plist"], thermo_defaults, lx=npoints))

    # Delete processed keys
    for key in thermo_keys:
        if key in sys_dict:
            del sys_dict[key]

    if "MultiprocessingObject" in sys_dict:
        MultiprocessingObject = sys_dict["MultiprocessingObject"]
        del sys_dict["MultiprocessingObject"]
        flag_use_mp_object = True
    else:
        flag_use_mp_object = False

    opts = sys_dict.copy()

    inputs = [
        (
            thermo_dict["Plist"][i],
            thermo_dict["Tlist"][i],
            thermo_dict["xilist"][i],
            Eos,
            opts,
        )
        for i in range(npoints)
    ]
    if flag_use_mp_object:
        (
            residual_v1,
            residual_v2,
            flagv,
            log_phiv,
            residual_l1,
            residual_l2,
            flagl,
            log_phil,
        ) = MultiprocessingObject.pool_job(_verify_eos_wrapper, inputs)
    else:
        (
            residual_v1,
            residual_v2,
            flagv,
            log_phiv,
            residual_l1,
            residual_l2,
            flagl,
            log_phil,
        ) = MultiprocessingJob.serial_job(_verify_eos_wrapper, inputs)

    logger.info("--- Calculation verify_eos Complete ---")

    return {
        "P": thermo_dict["Plist"],
        "T": thermo_dict["Tlist"],
        "xi": thermo_dict["xilist"],
        "residual_v1": residual_v1,
        "residual_v2": residual_v2,
        "flagv": flagv,
        "log_phivi": log_phiv,
        "residual_l1": residual_l1,
        "residual_l2": residual_l2,
        "flagl": flagl,
        "log_phili": log_phil,
    }


def _verify_eos_wrapper(args):
    r"""Wrapper for parallelized use of 'verify_eos' calculation type."""

    P, T, xi, Eos, opts = args

    logger.info("T (K), P (Pa), xi: {} {} {}, Let's Begin!".format(T, P, xi))

    rhov, flagv = calc.calc_vapor_density(P, T, xi, Eos, **opts)
    if np.isnan(rhov):
        logger.warning("Failed to calculate rhov at {} K and {} Pa".format(T, P))
        log_phiv, residual_v1, residual_v2 = np.nan, np.nan, np.nan
        logger.debug("Calculation Failed:", exc_info=True)
    else:
        phiv = Eos.fugacity_coefficient(P, rhov, xi, T)
        log_phiv = np.log(phiv)
        residual_v1 = calc.fugacity_test_1(P, T, xi, rhov, Eos, **opts)
        residual_v2 = calc.fugacity_test_2(P, T, xi, rhov, Eos, **opts)
        logger.info(
            "rhov {}, flagv {}, ".format(
                rhov,
                flagv,
            )
            + "log_phiv {}, log_phiv {}, residual1 {}, residual2 {}".format(
                np.sum(xi * log_phiv), log_phiv, residual_v1, residual_v2
            )
        )

    rhol, flagl = calc.calc_liquid_density(P, T, xi, Eos, **opts)
    if np.isnan(rhol):
        logger.warning("Failed to calculate rhol at {} K and {} Pa".format(T, P))
        log_phil, residual_l1, residual_l2 = np.nan, np.nan, np.nan
        logger.debug("Calculation Failed:", exc_info=True)
    else:
        phil = Eos.fugacity_coefficient(P, rhol, xi, T)
        log_phil = np.log(phil)
        residual_l1 = calc.fugacity_test_1(P, T, xi, rhol, Eos, **opts)
        residual_l2 = calc.fugacity_test_2(P, T, xi, rhol, Eos, **opts)
        logger.info(
            "rhol {}, flagl {}, ".format(rhol, flagl)
            + "log_phil {}, log_phil {}, residual1 {}, residual2 {}".format(
                np.sum(xi * log_phil), log_phil, residual_l1, residual_l2
            )
        )

    return (
        residual_v1,
        residual_v2,
        flagv,
        log_phiv,
        residual_l1,
        residual_l2,
        flagl,
        log_phil,
    )



""" General functions that can be used by multiple modules
"""

import numpy as np
import scipy.optimize as spo
import logging

logger = logging.getLogger(__name__)


def solve_root(func, args=(), method="bisect", x0=None, bounds=None, options={}):
    """
    This function will setup and dispatch thermodynamic jobs.

    Parameters
    ----------
    func : function
        Function used in job. Can be any of the following scipy methods: "brent",
        "least_squares", "TNC", "L-BFGS-B", "SLSQP", 'hybr', 'lm', 'linearmixing',
        'diagbroyden', 'excitingmixing', 'krylov', 'df-sane', 'anderson',
        'hybr_broyden1', 'hybr_broyden2', 'broyden1', 'broyden2', 'bisect'.
    args : tuple, Optional, default=()
        Each entry of this list contains the input arguments for each job
    method : str, Optional, default="bisect"
        Choose the method used to solve the dew point calculation
    x0 : float, Optional, default=None
        Initial guess in parameter to be optimized
    bounds : tuple, Optional, default=None
        Parameter boundaries
    options : dict, Optional, default={}
        These options are used in the scipy method

    Returns
    -------
    output : tuple
        This structure contains the outputs of the jobs given

    """

    if method not in [
        "brentq",
        "least_squares",
        "TNC",
        "L-BFGS-B",
        "SLSQP",
        "hybr",
        "lm",
        "linearmixing",
        "diagbroyden",
        "excitingmixing",
        "krylov",
        "df-sane",
        "anderson",
        "hybr_broyden1",
        "hybr_broyden2",
        "broyden1",
        "broyden2",
        "bisect",
    ]:
        raise ValueError("Optimization method, {}, not supported.".format(method))

    if x0 is None:
        logger.debug("Initial guess in optimization not provided")
    if np.any(bounds is None):
        logger.debug("Optimization bounds not provided")

    if x0 is None and method in [
        "broyden1",
        "broyden2",
        "anderson",
        "hybr",
        "lm",
        "linearmixing",
        "diagbroyden",
        "excitingmixing",
        "krylov",
        "df-sane",
    ]:
        if np.any(bounds is None):
            raise ValueError(
                "Optimization method, {}, requires x0. Because bounds were"
                " not provided, so problem cannot be solved.".format(method)
            )
        else:
            logger.error("Optimization method, {}, requires x0, using bisect instead".format(method))
            method = "bisect"

    if np.size(x0) > 1 and method in ["brentq", "bisect"]:
        logger.error("Optimization method, {}, is for scalar functions, using {}".format(method, "least_squares"))
        method = "least_squares"

    if not isiterable(bounds[0]):
        bounds = [bounds]

    if np.any([np.any(x is None) for x in bounds]) and method in ["brentq", "bisect"]:
        if x0 is None:
            raise ValueError(
                "Optimization method, {}, requires bounds. ".format(method)
                + "Because x0 was not provided, so problem cannot be solved."
            )
        else:
            logger.error("Optimization method, {}, requires bounds, using hybr".format(method))
            method = "hybr"

    if np.any(bounds is not None):
        for i, bnd in enumerate(bounds):
            if len(bnd) != 2:
                raise ValueError("bounds are not of length two")
            else:
                bounds[i] = tuple(bnd)

    # ################### Root Finding without Boundaries ###################
    if method in ["broyden1", "broyden2"]:
        outer_dict = {
            "fatol": 1e-5,
            "maxiter": 25,
            "jac_options": {"reduction_method": "simple"},
        }
        for key, value in options.items():
            outer_dict[key] = value
        logger.debug("Using the method, {}, with the following options:\n{}".format(method, outer_dict))
        sol = spo.root(func, x0, args=args, method=method, options=outer_dict)
    elif method == "anderson":
        outer_dict = {"fatol": 1e-5, "maxiter": 25}
        for key, value in options.items():
            outer_dict[key] = value
        logger.debug("Using the method, {}, with the following options:\n{}".format(method, outer_dict))
        sol = spo.root(func, x0, args=args, method=method, options=outer_dict)
    elif method in [
        "hybr",
        "lm",
        "linearmixing",
        "diagbroyden",
        "excitingmixing",
        "krylov",
        "df-sane",
    ]:
        outer_dict = {}
        for key, value in options.items():
            outer_dict[key] = value
        logger.debug("Using the method, {}, with the following options:\n{}".format(method, outer_dict))
        sol = spo.root(func, x0, args=args, method=method, options=outer_dict)

    # ################### Minimization Methods with Boundaries ###################
    elif method in ["TNC", "L-BFGS-B"]:
        outer_dict = {
            "gtol": 1e-2 * np.sqrt(np.finfo("float").eps),
            "ftol": np.sqrt(np.finfo("float").eps),
        }
        for key, value in options.items():
            outer_dict[key] = value
        logger.debug("Using the method, {}, with the following options:\n{}".format(method, outer_dict))
        if len(bounds) == 2:
            sol = spo.minimize(
                func,
                x0,
                args=args,
                method=method,
                bounds=tuple(bounds),
                options=outer_dict,
            )
        else:
            sol = spo.minimize(func, x0, args=args, method=method, options=outer_dict)
    elif method == "SLSQP":
        outer_dict = {}
        for key, value in options.items():
            outer_dict[key] = value
        logger.debug("Using the method, {}, with the following options:\n{}".format(method, outer_dict))
        if len(bounds) == 2:
            sol = spo.minimize(
                func,
                x0,
                args=args,
                method=method,
                bounds=tuple(bounds),
                options=outer_dict,
            )
        else:
            sol = spo.minimize(func, x0, args=args, method=method, options=outer_dict)

    # ################### Root Finding with Boundaries ###################
    elif method == "brentq":
        outer_dict = {"rtol": 1e-7}
        for key, value in options.items():
            if key in ["xtol", "rtol", "maxiter", "full_output", "disp"]:
                outer_dict[key] = value
        logger.debug("Using the method, {}, with the following options:\n{}".format(method, outer_dict))
        sol = spo.brentq(func, bounds[0][0], bounds[0][1], args=args, **outer_dict)
    elif method == "least_squares":
        outer_dict = {}
        for key, value in options.items():
            outer_dict[key] = value
        logger.debug("Using the method, {}, with the following options:\n{}".format(method, outer_dict))
        bnd_tmp = [[], []]
        for bnd in bounds:
            bnd_tmp[0].append(bnd[0])
            bnd_tmp[1].append(bnd[1])
        sol = spo.least_squares(func, x0, bounds=tuple(bnd_tmp), args=args, **outer_dict)
    elif method == "bisect":
        outer_dict = {"maxiter": 100, "rtol": 1e-12}
        for key, value in options.items():
            if key in ["xtol", "rtol", "maxiter", "full_output", "disp"]:
                outer_dict[key] = value
        logger.debug("Using the method, {}, with the following options:\n{}".format(method, outer_dict))
        sol = spo.bisect(func, bounds[0][0], bounds[0][1], args=args, **outer_dict)

    # Given final P estimate
    if method not in ["brentq", "bisect"]:
        solution = sol.x
        logger.info("Optimization terminated successfully: {} {}".format(sol.success, sol.message))
    else:
        logger.info("Optimization terminated successfully: {}".format(sol))
        solution = sol

    return solution


def central_difference(x, func, step_size=1e-5, relative=False, args=()):
    """
    Take the derivative of a dependent variable calculated with a given function using
    the central difference method.

    Parameters
    ----------
    x : numpy.ndarray
        Independent variable to take derivative with respect too, using the central
        difference method. Must be first input of the function.
    func : function
        Function used in job to calculate dependent factor. This function should have
        a single output.
    step_size : float, Optional, default=1E-5
        Either the step size used in the central difference method, or if
        ``relative=True``, this variable is a scaling factor so that the step size for
        each value of x is x * step_size.
    args : tuple, Optional, default=()
        Each entry of this list contains the input arguments for each job
    relative : bool, Optional, default=False
        If False, the step_size is directly used to calculate the derivative. If true,
        step_size becomes a scaling factor, where the step size for each value of x
        becomes step_size*x.

    Returns
    -------
    dydx : numpy.ndarray
        Array of derivative of y with respect to x, given an array of independent
        variables.
    """

    if not isiterable(x):
        x - np.array([x])
    elif not isinstance(x, np.ndarray):
        x = np.array(x)

    if relative:
        step = x * step_size
        if not isiterable(step):
            step = np.array([step])
        step = np.array([2 * np.finfo(float).eps if xx < np.finfo(float).eps else xx for xx in step])
    else:
        step = step_size

    y = func(np.append(x + step, x - step), *args)
    lx = int(len(y) / 2)
    dydx = (y[:lx] - y[lx:]) / (2.0 * step)

    return dydx


def isiterable(array):
    """
    Check if variable is an iterable type with a length (e.g. np.array or list).

    Note that this could be tested with ``isinstance(array, Iterable)``, however
    ``array=np.array(1.0)`` would pass that test and then fail in ``len(array)``.

    Parameters
    ----------
    array
        Variable of some type, that should be iterable

    Returns
    -------
    isiterable : bool
        Will be True if indexing is possible and False if not.
    """

    array_tmp = np.array(array, dtype=object)
    tmp = np.shape(array_tmp)
    if tmp:
        isiterable = True
    else:
        isiterable = False

    return isiterable


def check_length_dict(dictionary, keys, lx=None):
    """
    This function compared the entries, keys, in the provided dictionary to ensure
    they're the same length.

    All entries in the list ``keys``, will be made into numpy arrays (if present).
    If a float or array of length one is provided, it will be expanded to the length
    of other arrays.

    Parameters
    ----------
    dictionary : dict
        Dictionary containing all or some of the keywords, ``keys``, of what should
        be arrays of identical size.
    keys : list
        Possible keywords representing array entries
    lx : int, Optional, default=None
        The size that arrays should conform to

    Returns
    -------
    new_dictionary : dict
        Dictionary of arrays of identical size.

    """

    if lx is None:
        lx_array = []
        for key in keys:
            if key in dictionary:
                tmp = dictionary[key]
                if np.shape(tmp):
                    lx_array.append(len(tmp))
                else:
                    lx_array.append(1)
        if not len(lx_array):
            raise ValueError("None of the provided keys are found in the given dictionary")
        lx = max(lx_array)

    new_dictionary = {}
    for key in keys:
        if key in dictionary:
            tmp = dictionary[key]
            if isiterable(tmp):
                l_tmp = len(tmp)
                if l_tmp == 1:
                    new_dictionary[key] = np.array([tmp[0] for x in range(lx)], float)
                elif l_tmp == lx:
                    new_dictionary[key] = np.array(tmp, float)
                else:
                    raise ValueError("Entry, {}, should be length {}, not {}".format(key, lx, l_tmp))
            else:
                new_dictionary[key] = np.array([tmp for x in range(lx)], float)

    return new_dictionary


def set_defaults(dictionary, keys, values, lx=None):
    """
    This function checks a dictionary for the given keys, and if it's not there, the
    appropriate value is added to the dictionary.

    Parameters
    ----------
    dictionary : dict
        Dictionary of data
    keys : list
        Keys that should be present (of the same length as ``lx``)
    values : list
        Default values for the keys that aren't in dictionary
    lx : int, Optional, default=None
        If not None, and values[i] is a float, the key will be set to an array of
        length, ``lx``, populated by ``values[i]``

    Returns
    -------
    new_dictionary : dict
        Dictionary of arrays of identical size.

    """

    new_dictionary = dictionary.copy()

    key_iterable = isiterable(keys)
    if not isiterable(values):
        if key_iterable:
            values = np.ones(len(keys)) * values
        else:
            values = np.array([values])
            keys = np.array([keys])
    else:
        if key_iterable and len(keys) != len(values):
            raise ValueError("Length of given keys and values must be equivalent.")
        elif not key_iterable:
            if len(values) != 1:
                raise ValueError("Multiple default values for given key, {}, is ambiguous".format(keys))
            else:
                keys = [keys]

    for i, key in enumerate(keys):
        if key not in dictionary:
            tmp = values[i]
            if not isiterable(tmp) and lx is not None:
                new_dictionary[key] = np.ones(lx) * tmp
            else:
                new_dictionary[key] = tmp
            logger.info("Entry, {}, set to default: {}".format(key, tmp))

    return new_dictionary



"""
Unit and regression test for the despasito package.
"""

# Import package, test suite, and other packages as needed
import despasito
import logging
import random
import os

logger = logging.getLogger(__name__)


def test_despasito_log_file():
    """Test enabling of logging"""

    fname = "despasito_{}.log".format(random.randint(1, 10))
    print(fname)
    despasito.initiate_logger(log_file=fname, verbose=10)
    logger.info("test")

    if os.path.isfile(fname):
        flag = True
        despasito.initiate_logger(log_file=False)
        try:
            os.remove(fname)
        except Exception:
            print("Error removing log file")
    else:
        flag = False

    assert flag


def test_despasito_log_console(capsys):
    """Test enabling of logging"""

    despasito.initiate_logger(console=True, verbose=10)
    logger.info("test")

    _, err = capsys.readouterr()

    despasito.initiate_logger(console=False)

    assert "[INFO](despasito.tests.test_logging): test" in err



# -- coding: utf8 --
r"""
    EOS object for SAFT-:math:`\gamma`-SW

    Equations referenced in this code are from Lymperiadis, A. et. al, J. Chem. Phys.
    127, 234903 (2007)
"""

import numpy as np
import logging

import despasito.equations_of_state.eos_toolbox as tb
from despasito.equations_of_state import constants
import despasito.equations_of_state.saft.saft_toolbox as stb
from despasito.equations_of_state.saft import Aassoc
from despasito.equations_of_state import method_stat

logger = logging.getLogger(__name__)


if not method_stat.cython and not method_stat.numba:
    pass
elif method_stat.cython:
    logger.warning("saft.gamma_sw does not use cython.")
elif method_stat.numba:
    logger.warning("saft.gamma_sw does not use numba.")

ckl_coef = np.array(
    [
        [2.25855, -1.50349, 0.249434],
        [-0.669270, 1.40049, -0.827739],
        [10.1576, -15.0427, 5.30827],
    ]
)


class SaftType:
    r"""
    Object of SAFT variant

    Parameters
    ----------
    beads : list[str]
        List of unique bead names used among components
    molecular_composition : numpy.ndarray
        :math:`\\nu_{i,k}/k_B`. Array of number of components by number of bead types.
        Defines the number of each type of group in each component.
    bead_library : dict
        A dictionary where bead names are the keys to access EOS self interaction
        parameters
    cross_library : dict, Optional, default={}
        Optional library of bead cross interaction parameters. As many or as few of the
        desired parameters may be defined for whichever group combinations are desired.
        If this matrix isn't provided, the SAFT mixing rules are used.

    Attributes
    ----------
    beads : list[str]
        List of unique bead names used among components
    bead_library : dict
        A dictionary where bead names are the keys to access EOS self interaction
        parameters. See **Parameters** section.
    cross_library : dict
        Library of bead cross interaction parameters. As many or as few of the desired
        parameters may be defined for whichever group combinations are desired. See
        **Parameters** section.
    Aideal_method : str
        The default functional form of the ideal gas contribution of the Helmholtz
        energy
    residual_helmholtz_contributions : list[str]
        List of methods from the specified saft_source representing contributions to
        the Helmholtz energy that are functions of density, temperature, and
        composition.
    parameter_types : list[str]
        This list of parameter names for the specific SAFT variant.
    parameter_bound_extreme : dict
        With each parameter name as an entry representing a list with the minimum and
        maximum feasible parameter value.
    combining_rules : dict
        Contains functional form and additional information for calculating cross
        interaction parameters that are not found in `cross_library`. Function must be
        one of those contained in
        :mod:`~despasito.equations_of_state.combining_rule_types`.
    eos_dict : dict
        Dictionary of parameters and specific settings

        - molecular_composition (numpy.ndarray) - :math:`\\nu_{i,k}/k_B`. Array of
        number of components by number of bead types. Defines the number of each type
        of group in each component.

    ncomp : int
        Number of components in the system
    nbeads : int
        Number of beads in system that are shared among components

    """

    def __init__(self, **kwargs):

        # ###### The Following lines are **MANDATORY** but can be edited as needed

        # Standard attributes of the EosTemplate are 'density_max' and
        # 'parameter_refresh', while all saft versions require 'calc_gr_assoc' for
        # calculating association sites

        # This keyword is used by the saft.py class and subsequently the
        # `Aideal_contribution` function. If a user desired a different default method
        # of calculating the ideal contribution, add a new function to Aideal.py
        self.Aideal_method = "Abroglie"

        # This list of strings are the bead types that the parameter update method can
        # expect.
        # The bead library should use these same parameter names.
        # Parameter names should not contain '-' or '_'
        self.parameter_types = ["example"]

        # This dictionary will contain the feasible bounds of the parameters listed in
        # 'parameter_types'. Bounds are necessary for each type.
        self.parameter_bound_extreme = {"example": [0, 1]}

        # When calculating the cross-interaction term between two beads, the parameter
        # name and the combining rule type should be listed here. The combining rule
        # keyword can be any that are supported in saft_toolbox.combining_rules. This
        # attribute must exist.
        self.combining_rules = {"example": {"function": "mean"}}

        # This list contains the Helmholtz energy contributions contained in this class
        # below. The class in saft.py will add these as it's own attributes to
        # calculate the total Helmholtz energy.
        self.residual_helmholtz_contributions = ["Amonomer", "Achain"]
        # Note that these strings must represent methods for the Helmholtz
        # contribution below and are added to Aideal and Aassoc in the main SAFT class.
        # If a function is to be optimized with Cython or Numba, an extensions library
        # can be added to the 'compiled_modules' directory. A python version with the
        # same function names should also be present as well. The import if-structure
        # at the beginning of this module is then used to import the desired form, see
        # gamma_mie.py for an example.
        # When deciding whether to include an additional function as a class method, or
        # if it should simply be imported from a library. Think about whether accessing
        # that function would be nice when handling an EOS object in a python script.
        # For instance, the 'reduced_density', 'effective_packing_fraction',
        # 'Ahard_sphere', 'Afirst_order', and other methods in gamma_mie.py would be
        # nice to have as attributes.

        # Now we start processing the given variables. The following three attributes
        # are always needs for the saft.py class. If other inputs are needed for the
        # specific SAFT type at hand, feel free to add them to this list.
        if not hasattr(self, "eos_dict"):
            self.eos_dict = {}

        needed_attributes = ["molecular_composition", "beads", "bead_library"]
        for key in needed_attributes:
            if key not in kwargs:
                raise ValueError("The one of the following inputs is missing: {}".format(", ".join(needed_attributes)))
            elif key == "molecular_composition":
                if "molecular_composition" not in self.eos_dict:
                    self.eos_dict[key] = kwargs[key]
            elif not hasattr(self, key):
                setattr(self, key, kwargs[key])

        # Check bead_library to be sure all parameters are present. If one is missing
        # that has a standard default then that is added
        self._parameter_defaults = {
            "epsilon": None,
            "lambdar": None,
            "lambdaa": None,
            "sigma": None,
            "Sk": 1.0,
            "Vks": 1.0,
        }
        self.bead_library = tb.check_bead_parameters(self.bead_library, self._parameter_defaults)

        if "cross_library" not in kwargs:
            self.cross_library = {}
        else:
            self.cross_library = kwargs["cross_library"]

        # ##### The following lines are *OPTIONAL* and are completely for internal use
        # for this specific SAFT type and aren't mandatory

        # Initialize composition attribute. This is for composition dependent
        # properties. By recording this, we can avoid recalculating those
        # parameters unnecessarily. In saft-gamma_mie we also have temperature
        # dependent parameters and so self.T is included.
        if not hasattr(self, "xi"):
            self.xi = np.nan

        # These are initialized for loops used in some of the methods below. Depending
        # on how you choose to break things up, these might not be needed.
        if not hasattr(self, "nbeads") or not hasattr(self, "ncomp"):
            self.ncomp, self.nbeads = np.shape(self.eos_dict["molecular_composition"])

        # Initiate cross interaction terms, as mentioned above, some combining rules use
        # a particular combination of parameters and these are handled here.
        output = tb.cross_interaction_from_dict(
            self.beads,
            self.bead_library,
            self.combining_rules,
            cross_library=self.cross_library,
        )
        self.eos_dict["sigma_kl"] = output["sigma"]
        self.eos_dict["epsilon_kl"] = output["epsilon"]
        self.eos_dict["lambda_kl"] = output["lambda"]

        # This optional keyword can be passed in an input file as "eos_num_rings". If
        # your chosen version of SAFT needs special keyword passed that don't fall
        # under a category above. Make a note in the doc string to pass it as
        # "eos_somekeyword"
        if "num_rings" in kwargs:
            self.eos_dict["num_rings"] = kwargs["num_rings"]
            logger.info("Accepted component ring structure: {}".format(kwargs["num_rings"]))
        else:
            self.eos_dict["num_rings"] = np.zeros(len(self.eos_dict["molecular_composition"]))

    def Amonomer(self, rho, T, xi):
        r"""

        ** Example of Helmholtz contribution to put in
        residual_helmholtz_contributions **

        Outputs the monomer contribution of the Helmholtz energy
        :math:`A^{mono.}/Nk_{b}T`.

        Parameters
        ----------
        rho : numpy.ndarray
            Number density of system [mol/m^3]
        T : float
            Temperature of the system [K]
        xi : numpy.ndarray
            Mole fraction of each component, sum(xi) should equal 1.0

        Returns
        -------
        Amonomer : numpy.ndarray
            Helmholtz energy of monomers for each density given.
        """

        if np.all(rho > self.density_max(xi, T)):
            raise ValueError(
                "Density values should not all be greater than "
                + "{}, or calc_Amono".format(self.density_max(xi, T))
                + " will fail in log calculation."
            )

        rho = self._check_density(rho)
        self._check_composition_dependent_parameters(xi)

        zetax = self.reduced_density(rho, xi)[:, 3]

        Amonomer = (
            self.Ahard_sphere(rho, T, xi)
            + self.Afirst_order(rho, T, xi, zetax=zetax)
            + self.Asecond_order(rho, T, xi, zetax=zetax)
        )

        return Amonomer

    def density_max(self, xi, T, maxpack=0.65):
        """
        ** Mandatory **

        Estimate the maximum density based on the hard sphere packing fraction.

        Parameters
        ----------
        xi : list[float]
            Mole fraction of each component
        T : float
            Temperature of the system [K]
        maxpack : float, Optional, default=0.65
            Maximum packing fraction

        Returns
        -------
        max_density : float
            Maximum molar density [mol/m^3]
        """

        self._check_composition_dependent_parameters(xi)

        # estimate the maximum density based on the hard sphere packing fraction
        # etax, assuming a maximum packing fraction specified by maxpack
        max_density = (
            maxpack
            * 6.0
            / (self.eos_dict["Cmol2seg"] * np.pi * np.sum(self.eos_dict["xskl"] * (self.eos_dict["sigma_kl"] ** 3)))
            / constants.molecule_per_nm3
        )

        return max_density

    def calc_gr_assoc(self, rho, T, xi, Ktype="ijklab"):
        r"""
        ** Mandatory **

        Reference fluid pair correlation function used in calculating association
        sites

        Parameters
        ----------
        rho : numpy.ndarray
            Number density of system [mol/m^3]
        T : float
            Temperature of the system [K]
        xi : numpy.ndarray
            Mole fraction of each component, sum(xi) should equal 1.0
        Ktype : str, Optional, default='ijklab'
            Indicates which radial distribution function to return. The only option is
            'ijklab': The bonding volume was calculated from self.calc_Kijklab, return
            gHS_dij)

        Returns
        -------
        gr : numpy.ndarray
            A temperature-density polynomial correlation of the association integral
            for a Lennard-Jones monomer. This matrix is (len(rho) x Ncomp x Ncomp)
        """

        rho = self._check_density(rho)
        gSW = self.calc_gSW(rho, T, xi)

        return gSW

    def calc_Kijklab(self, T, rc_klab, rd_klab=None, reduction_ratio=0.25):
        r"""
        ** Mandatory **

        Calculation of association site bonding volume, dependent on molecule in
        addition to group

        Lymperiadis Fluid Phase Equilibria 274 (2008) 85–104

        Parameters
        ----------
        T : float
            Temperature of the system [K], Note used in this version of saft, but
            included to allow saft.py to be general
        rc_klab : numpy.ndarray
            This matrix of cutoff distances for association sites for each site type in
            each group type
        rd_klab : numpy.ndarray, Optional, default=None
            Position of association site in each group (nbead, nbead, nsite, nsite)
        reduction_ratio : float, Optional, default=0.25
            Reduced distance of the sites from the center of the sphere of interaction.
            This value is used when site position, rd_klab is None

        Returns
        -------
        Kijklab : numpy.ndarray
            Matrix of binding volumes
        """

        dij_bar = np.zeros((self.ncomp, self.ncomp))
        for i in range(self.ncomp):
            for j in range(self.ncomp):
                dij_bar[i, j] = np.mean([self.eos_dict["sigma_ij"][i], self.eos_dict["sigma_ij"][j]])

        Kijklab = Aassoc.calc_bonding_volume(rc_klab, dij_bar, rd_klab=rd_klab, reduction_ratio=reduction_ratio)

        return Kijklab

    def parameter_refresh(self, bead_library, cross_library):
        r"""
        ** Mandatory **

        To refresh dependent parameters

        Those parameters that are dependent on bead_library and cross_library attributes
        **must** be updated by running this function after all parameters from
        update_parameters method have been changed.

        Attributes
        ----------
        alpha : np.array
            van der Waals attractive parameter for square-well segments, equal to
            :math:`\alpha_{k,l}/k_B`.
        eos_dict : dict
            The following entries are updated:

            - epsilon_kl (numpy.ndarray) - Matrix of well depths for groups (k,l)
            - sigma_kl (numpy.ndarray) - Matrix of bead diameters (k,l)
            - lambda_kl (numpy.ndarray) - Matrix of range of potential well
            depth (k,l)
            - xskl (numpy.ndarray) - Matrix of mole fractions of bead (i.e. segment or
            group) k multiplied by that of bead l
            - Cmol2seg (float) - Conversion factor from from molecular number density,
            :math:`\rho`, to segment (i.e. group) number density, :math:`\rho_S`.

        """

        self.bead_library.update(bead_library)
        self.cross_library.update(cross_library)

        # Update Non bonded matrices
        output = tb.cross_interaction_from_dict(
            self.beads,
            self.bead_library,
            self.combining_rules,
            cross_library=self.cross_library,
        )
        self.eos_dict["sigma_kl"] = output["sigma"]
        self.eos_dict["epsilon_kl"] = output["epsilon"]
        self.eos_dict["lambda_kl"] = output["lambda"]
        self.calc_component_averaged_properties()

        if not np.any(np.isnan(self.xi)):
            self.eos_dict["Cmol2seg"], self.eos_dict["xskl"] = stb.calc_composition_dependent_variables(
                self.xi,
                self.eos_dict["molecular_composition"],
                self.bead_library,
                self.beads,
            )
        self.alphakl = (
            2.0
            * np.pi
            / 3.0
            * self.eos_dict["epsilon_kl"]
            * self.eos_dict["sigma_kl"] ** 3
            * (self.eos_dict["lambda_kl"] ** 3 - 1.0)
        )

    def _check_density(self, rho):
        r"""
        ** Mandatory **

        This function checks that the density array is in the correct format for
        further calculations.

        Parameters
        ----------
        rho : numpy.ndarray
            Number density of system [mol/m^3]

        Returns
        -------
        rho : numpy.ndarray
            Number density of system [mol/m^3]
        """

        if np.isscalar(rho):
            rho = np.array([rho])
        elif not isinstance(rho, np.ndarray):
            rho = np.array(rho)
        if len(np.shape(rho)) == 2:
            rho = rho[0]

        if any(np.isnan(rho)):
            raise ValueError("NaN was given as a value of density, rho")
        elif rho.size == 0:
            raise ValueError("No value of density was given")
        elif any(rho < 0.0):
            raise ValueError("Density values cannot be negative.")

        return rho



import numpy as np
import logging
from inspect import getmembers, isfunction

from . import combining_rule_types

logger = logging.getLogger(__name__)


def remove_insignificant_components(xi_old, massi_old):
    r"""
    This function will remove any components with mole fractions less than or equal
    to machine precision.

    Parameters
    ----------
    xi_old : numpy.ndarray
        Mole fraction of each component, sum(xi) should equal 1.0
    massi_old : numpy.ndarray
        Mass for each component [kg/mol]

    Returns
    -------
    xi_new : numpy.ndarray
        Mole fraction of each component, sum(xi) should equal 1.0
    massi_new : numpy.ndarray
        Mass for each component [kg/mol]

    """
    ind = np.where(np.array(xi_old) < np.finfo(float).eps)[0]
    xi_new = []
    massi_new = []
    for i in range(len(xi_old)):
        if i not in ind:
            xi_new.append(xi_old[i])
            massi_new.append(massi_old[i])
    xi_new = np.array(xi_new)
    massi_new = np.array(massi_new)

    return xi_new, massi_new


def partial_density_central_difference(xi, rho, T, func, step_size=1e-2, log_method=False):
    """
    Take the derivative of a dependent variable calculated with a given function
    using the central difference method.

    Parameters
    ----------
    xi : list[float]
        Mole fraction of each component
    rho : float
        Molar density of system [:math:`mol/m^3`]
    T : float
        Temperature of the system [K]
    func : function
        Function used in job to calculate dependent factor. This function should have
        a single output. Inputs arguments should be (rho, T, xi)
    step_size : float, Optional, default=1E-2
        Step size used in central difference method
    log_method : bool, Optional, default=False
        Choose to use a log transform in central difference method. This allows easier
        calculations for very small numbers.

    Returns
    -------
    dydxi : numpy.ndarray
        Array of derivative of y with respect to xi
    """

    dAdrho = np.zeros(len(xi))

    if log_method:  # Central Difference Method with log(y) transform

        dy = step_size
        y = np.log(rho * np.array(xi, float))
        for i in range(np.size(dAdrho)):
            if xi[i] != 0.0:
                Ares = np.zeros(2)
                for j, delta in enumerate((dy, -dy)):
                    y_temp = np.copy(y)
                    y_temp[i] += delta
                    Ares[j] = _partial_density_wrapper(np.exp(y_temp), T, func)
                dAdrho[i] = (Ares[0] - Ares[1]) / (2.0 * dy) / np.exp(y[i])
            else:
                dAdrho[i] = np.finfo(float).eps

    else:  # Traditional Central Difference Method

        dy = step_size
        y = rho * np.array(xi, float)
        for i in range(np.size(dAdrho)):
            if xi[i] != 0.0:
                Ares = np.zeros(2)
                for j, delta in enumerate((dy, -dy)):
                    y_temp = np.copy(y)
                    if y_temp[i] != 0.0:
                        y_temp[i] += delta
                    Ares[j] = _partial_density_wrapper(y_temp, T, func)
                dAdrho[i] = (Ares[0] - Ares[1]) / (2.0 * dy)
            else:
                dAdrho[i] = np.finfo(float).eps

    return dAdrho


def _partial_density_wrapper(rhoi, T, func):
    """
    Compute derivative of Helmholtz energy with respect to density.

    Parameters
    ----------
    rhoi : float
        Molar density of each component, add up to the total density [:math:`mol/m^3`]
    T : float
        Temperature of the system [K]
    func : function
        Function used in job to calculate dependent factor. This function should have
        a single output.

    Returns
    -------
    Ares : float
        Helmholtz energy give number of moles, length of array rho
    """

    # Calculate new xi values
    rho = np.array([np.sum(rhoi)])
    xi = rhoi / rho

    Ares = func(rho, T, xi)[0]

    return Ares


def calc_massi(molecular_composition, bead_library, beads):
    r"""
    This function extracted the mass of each component

    Parameters
    ----------
    molecular_composition : numpy.ndarray
        :math:`\nu_{i,k}/k_B`. Array of number of components by number of bead types.
        Defines the number of each type of group in each component.
    bead_library : dict
        A dictionary where bead names are the keys to access EOS self interaction
        parameters:

        - mass: Bead mass [kg/mol]

    beads : list[str]
        List of unique bead names used among components

    Returns
    -------
    massi : numpy.ndarray
        Bead mass corresponding to array ``beads`` [kg/mol]
    """
    massi = np.zeros(len(molecular_composition))
    for i in range(len(molecular_composition)):
        for k, bead in enumerate(beads):
            if "mass" in bead_library[bead]:
                massi[i] += molecular_composition[i, k] * bead_library[bead]["mass"]
            else:
                raise ValueError("The mass for bead, {}, was not provided.".format(bead))

    return massi


def extract_property(prop, bead_library, beads, default=None):
    r"""
    Extract single property or key from a dictionary within a dictionary (e.g. bead
    parameters) and into a single array of the same length and order as a list of bead
    names.

    The expected structure is a dictionary of dictionaries, such as a parameter
    library.

    Parameters
    ----------
    property : str
        Name of property in bead_library
    bead_library : dict
        A dictionary where bead names are the keys to access EOS self interaction
        parameters:
    beads : list[str]
        List of unique bead names used among components
    default : any, Optional, default=None
        If property is not present, set to this value. Although if the default is None,
        an error will result.

    Returns
    -------
    prop_array : numpy.ndarray
        array of desired property
    """

    prop_array = np.zeros(len(beads))
    for i, bead in enumerate(beads):
        if prop in bead_library[bead]:
            prop_array[i] = bead_library[bead][prop]
        else:
            if default is None:
                raise ValueError("The property {} for bead, {}, was not provided.".format(prop, bead))
            else:
                prop_array[i] = default

    return prop_array


def check_bead_parameters(bead_library0, parameter_defaults):
    r"""
    Be sure all needed parameters are available for each bead.

    If a parameter is absent and a default value is given, this value will be added to
    the parameter set. If the default is None, then an error is raised.

    Parameters
    ----------
    bead_library : dict
        A dictionary where bead names are the keys to access EOS self interaction
        parameters
    parameter_defaults : dict
        A dictionary of default values for the required parameters.

    Returns
    -------
    new_bead_library : dict
        New dictionary with defaults added where relevant

    """

    bead_library = bead_library0.copy()

    for bead, bead_dict in bead_library.items():
        for parameter, default in parameter_defaults.items():
            if parameter not in bead_dict:
                if default is not None:
                    bead_library[bead][parameter] = default
                    logger.info(
                        "Parameter, {}, is missing for parametrized group, {}. Set to "
                        "default, {}".format(parameter, bead, default)
                    )
                else:
                    raise ValueError(
                        "Parameter, {}, should have been defined for parametrized " "group, {}.".format(parameter, bead)
                    )

    return bead_library


def cross_interaction_from_dict(beads, bead_library, combining_dict, cross_library={}):
    r"""
    Computes matrices of cross interaction parameters defined as the keys in the
    combining dict parameter are extracted from the bead_library and then the cross
    library.

    Parameters
    ----------
    beads : list[str]
        List of unique bead names used among components
    bead_library : dict
        A dictionary where bead names are the keys to access EOS self interaction
        parameters. Those to be calculated are defined by the keys of combining_dict
    combining_dict : dict
        This dictionary contains those bead parameters that should be placed in a
        matrix and the combining rules for the cross parameters
    cross_library : dict, Optional, default={}
        Optional library of bead cross interaction parameters. As many or as few of the
        desired parameters may be defined for whichever group combinations are desired.
        If this matrix isn't provided, the SAFT combining rules are used.

    Returns
    -------
    output : dict
        Dictionary of outputs, with the same keys used in combining dict for the
        respective interaction matrix

    """

    nbeads = len(beads)

    # Set-up output dictionaries
    output = {}
    for key in combining_dict:
        output[key] = np.zeros((nbeads, nbeads))
        for k in range(nbeads):
            output[key][k, k] = bead_library[beads[k]][key]

    # Put cross_library in output matrices
    for i, beadname in enumerate(beads):
        for j, beadname2 in enumerate(beads):
            if j > i:
                for key in combining_dict:
                    if cross_library.get(beadname, {}).get(beadname2, {}).get(key, None) is not None:
                        output[key][i, j] = cross_library[beadname][beadname2][key]
                    elif cross_library.get(beadname2, {}).get(beadname, {}).get(key, None) is not None:
                        output[key][i, j] = cross_library[beadname2][beadname][key]
                    else:
                        try:
                            tmp = combining_rules(
                                bead_library[beadname], bead_library[beadname2], key, **combining_dict[key]
                            )
                        except Exception:
                            raise ValueError(
                                "Unable to calculate '{}' with '{}' method, for beads:"
                                " '{}' '{}'".format(
                                    key,
                                    combining_dict[key]["function"],
                                    beadname,
                                    beadname2,
                                )
                            )
                        for k2, v2 in tmp.items():
                            output[k2][i, j] = v2
                            output[k2][j, i] = v2
                    output[key][j, i] = output[key][i, j]

    return output


def construct_dummy_bead_library(input_dict, keys=None):
    r"""
    Using arrays of values, a dictionary is populated like a bead_library.

    If keys are included, they are used, otherwise, integers are used.

    Parameters
    ----------
    input_dict : dict
        A dictionary where parameter names are the keys to access EOS arrays of
        self-interaction parameters.
    keys : list[str], Optional, default=None
        List must be the same length as the lists in `input_dict`. These are used as
        labels.

    Returns
    -------
    output_dict : dict
        Dictionary of outputs, with the same keys used in combining dict for the
        respective interaction matrix

    """

    output = {}
    flag = False
    for parameter in input_dict:
        if keys is None:
            keys = [str(x) for x in range(len(input_dict[parameter]))]
            flag = True
        if len(keys) != len(input_dict[parameter]):
            raise ValueError("Number of keys is not equal to the number of quantities given")

        for i, bead in enumerate(keys):
            if bead not in output:
                output[bead] = {}
            output[bead][parameter] = input_dict[parameter][i]

    if flag:
        return output, keys
    else:
        return output


def combining_rules(beadA, beadB, parameter, function="mean", **kwargs):
    r"""
    Calculates cross interaction parameter according to the calculation method defined.

    Parameters
    ----------
    beadA : dict
        Dictionary of parameters used to describe a bead
    beadB : dict
        Dictionary of parameters used to describe a bead
    parameter : str
        Name of parameter for which a mixed value is needed
    function : str, Optional, default=mean
        Mixing rule function found in
        `despasito.equations_of_state.combining_rule_types.py`
    kwargs :
        Keyword arguments used in other averaging function

    Returns
    -------
    output_dict : dict
        Dictionary with keyword of parameter and mixed interaction parameter. If
        combining rule type outputs more than one updated variable, it will also be
        included
    """

    calc_list = [o[0] for o in getmembers(combining_rule_types) if isfunction(o[1])]
    try:
        if function != "None":
            func = getattr(combining_rule_types, function)
    except Exception:
        raise ImportError(
            "The combining rule type, '{}', was not found\nThe following calculation "
            "types are supported: {}".format(function, ", ".join(calc_list))
        )

    if function != "None":
        output = func(beadA, beadB, parameter, **kwargs)
        if not isinstance(output, dict):
            tmp = {parameter: output}
            output = tmp
    else:
        output = {}

    return output



r"""
Objects for storing and producing objective values for comparing experimental data
to EOS predictions.
"""

import numpy as np
import logging

from despasito.thermodynamics import thermo
from despasito.parameter_fitting import fit_functions as ff
from despasito.parameter_fitting.interface import ExpDataTemplate
import despasito.fundamental_constants as constants
import despasito.utils.general_toolbox as gtb

logger = logging.getLogger(__name__)


##################################################################
#                                                                #
#                       Liquid Density                           #
#                                                                #
##################################################################
class Data(ExpDataTemplate):
    r"""
    Object for liquid density data. This data is evaluated with "liquid_properties".

    This object is initiated in :func:`~despasito.parameter_fitting.fit` with the
    keyword, ``exp_data[*]["data_class_type"]="liquid_density"``.
    The data could be evaluated with
    :func:`~despasito.thermodynamics.calculation_types.liquid_properties`

    Parameters
    ----------
    data_dict : dict
        Dictionary of exp data of type liquid density.

        * calculation_type (str) - Optional, default='liquid_properties'
        * MultiprocessingObject (obj) - Optional, Initiated
        :class:`~despasito.utils.parallelization.MultiprocessingJob`
        * eos_obj (obj) - Equation of state object
        * T (list) - [K] List of temperature values for calculation
        * P (list) - [Pa] List of pressure values for calculation
        * xi (list) - List of liquid mole fractions used in liquid_properties
        calculations
        * rhol (list) - [mol/:math:`m^3`] Evaluated liquid density values
        * weights (dict) - A dictionary where each key is a system constraint
        (e.g. T or xi) which is also a header used in an optional exp. data file.
        The value associated with a header can be a list as long as the number of
        data points to multiply by the objective value associated with each point,
        or a float to multiply the objective value of this data set.
        * objective_method (str) - The 'method' keyword in function
        despasito.parameter_fitting.fit_functions.obj_function_form.
        * density_opts (dict) - Optional,
        default={"min_density_fraction":(1.0 / 60000.0), "density_increment":10.0,
        "max_volume_increment":1.0E-4}, Dictionary of options used in calculating
        pressure vs. mole fraction curves.
        * kwargs for
        :func:`~despasito.parameter_fitting.fit_functions.obj_function_form`

    Attributes
    ----------
    name : str
        Data type, in this case liquid_density
    Eos : obj
        Equation of state object
    weights : dict, Optional, default: {"some_property": 1.0 ...}
        Dictionary with keys corresponding to those in thermodict, with weighting
        factor or vector for each system property used in fitting
    obj_opts : dict
        Keywords to compute the objective function with
        :func:`~despasito.parameter_fitting.fit_functions.obj_function_form`.
    npoints : int
        Number of sets of system conditions this object computes
    result_keys : list
        Thermodynamic property names used in calculation of objective function. In in
        this case: ["rhol", "phil"]
    thermodict : dict
        Dictionary of inputs needed for thermodynamic calculations

        - calculation_type (str) default=liquid_properties
        - density_opts (dict) default={"min_density_fraction":(1.0 / 300000.0),
        "density_increment":10.0, "max_volume_increment":1.0E-4}

    """

    def __init__(self, data_dict):

        data_dict = data_dict.copy()

        super().__init__(data_dict)

        self.name = "liquid_density"
        tmp = {
            "min_density_fraction": (1.0 / 300000.0),
            "density_increment": 10.0,
            "max_volume_increment": 1.0e-4,
        }
        if "density_opts" in self.thermodict:
            tmp.update(self.thermodict["density_opts"])
        self.thermodict["density_opts"] = tmp

        if self.thermodict["calculation_type"] is None:
            self.thermodict["calculation_type"] = "liquid_properties"

        if "xi" in data_dict:
            self.thermodict["xilist"] = data_dict["xi"]
            del data_dict["xi"]
        if "T" in data_dict:
            self.thermodict["Tlist"] = data_dict["T"]
            del data_dict["T"]
        if "P" in data_dict:
            self.thermodict["Plist"] = data_dict["P"]
            del data_dict["P"]

        self.thermodict.update(data_dict)

        thermo_keys = ["Plist", "xilist", "Tlist"]
        self.result_keys = ["rhol", "phil"]

        key_list = list(set(thermo_keys + self.result_keys))
        self.thermodict.update(gtb.check_length_dict(self.thermodict, key_list))
        self.npoints = np.size(self.thermodict["Tlist"])

        if "xilist" not in self.thermodict and self.Eos.number_of_components > 1:
            raise ValueError("Ambiguous mixture composition. Define xi")
        thermo_defaults = [
            constants.standard_pressure,
            np.array([[1.0] for x in range(self.npoints)]),
            constants.standard_temperature,
        ]
        self.thermodict.update(gtb.set_defaults(self.thermodict, thermo_keys, thermo_defaults, lx=self.npoints))

        self.weights.update(gtb.check_length_dict(self.weights, self.result_keys, lx=self.npoints))
        self.weights.update(gtb.set_defaults(self.weights, self.result_keys, 1.0))

        if "Tlist" not in self.thermodict and "rhol" not in self.thermodict:
            raise ImportError("Given liquid property data, values for T, xi, and rhol should have" " been provided.")

        logger.info(
            "Data type 'liquid_properties' initiated with calculation_type, {}, and "
            "data types: {}.\nWeight data by: {}".format(
                self.thermodict["calculation_type"],
                ", ".join(self.result_keys),
                self.weights,
            )
        )

    def _thermo_wrapper(self):
        """
        Generate thermodynamic predictions from Eos object

        Returns
        -------
        phase_list : float
            A list of the predicted thermodynamic values estimated from thermo
            calculation. This list can be composed of lists or floats
        """

        # Remove results
        opts = self.thermodict.copy()
        tmp = self.result_keys + ["name", "parameters_guess"]
        for key in tmp:
            if key in opts:
                del opts[key]

        try:
            output_dict = thermo(self.Eos, **opts)
            output = [output_dict["rhol"]]
        except Exception:
            raise ValueError("Calculation of calc_liquid_density failed")
        return output

    def objective(self):
        """
        Generate objective function value from this dataset

        Returns
        -------
        obj_val : float
            A value for the objective function
        """

        phase_list = self._thermo_wrapper()

        # Reformat array of results
        phase_list, len_list = ff.reformat_output(phase_list)
        phase_list = np.transpose(np.array(phase_list))

        # objective function
        obj_value = ff.obj_function_form(
            phase_list, self.thermodict["rhol"], weights=self.weights["rhol"], **self.obj_opts
        )

        logger.info("Obj. breakdown for {}: rhol {}".format(self.name, obj_value))

        if np.isnan(obj_value) or obj_value == 0.0:
            obj_value = np.inf

        return obj_value



# -- coding: utf8 --

"""
Handles the primary functions

In any directory with the appropriate input files in the JSON format, run DESPASITO
with ``python -m despasito input.json``
"""

import logging
import argparse
import numpy as np

from despasito.input_output import read_input, write_output
from despasito.equations_of_state import initiate_eos
from despasito.thermodynamics import thermo
from despasito.parameter_fitting import fit

logger = logging.getLogger(__name__)


def get_parser():
    """Process line arguments"""

    # Define parser functions and arguments
    parser = argparse.ArgumentParser(
        description=(
            r"DESPASITO: Determining Equilibrium State and Parametrization "
            + "Application for SAFT, Intended for Thermodynamic Output.  This is an "
            + "open-source application for thermodynamic calculations and parameter"
            + " fitting for equations of state (EOS) like the Statistical "
            + "Associating Fluid Theory (SAFT) EOS."
        )
    )
    parser.add_argument(
        "-i",
        "--input",
        dest="input",
        help=(
            "Input file in JSON format with calculation instructions and path(s) to"
            + " equation of state parameters. See documentation for explicit "
            + "explanation. Compile docs or visit https://despasito.readthedocs.io"
        ),
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="count",
        default=0,
        help=("Verbose level: repeat up to three times for Warning, Info, or Debug " + "levels."),
    )
    parser.add_argument(
        "--log",
        nargs="?",
        dest="logFile",
        default="despasito.log",
        help="Name of output a log file.",
    )
    parser.add_argument(
        "-n",
        "--ncores",
        dest="ncores",
        type=int,
        help=("Set the number of cores used. A value of -1 will request all possible " + "resources."),
        default=1,
    )
    parser.add_argument(
        "-p",
        "--path",
        default=".",
        help=("Set the location of the data/library files (e.g. SAFTcross, etc.) where" + " despasito will look."),
    )
    parser.add_argument(
        "-c",
        "--console",
        action="store_true",
        help="Enable logs to print to standard output.",
    )
    parser.add_argument(
        "--numba",
        action="store_false",
        help="Turn on Numba's JIT compilation for accelerated computation",
    )
    parser.add_argument(
        "--python",
        action="store_true",
        help="Run calculation in python without compiled modules.",
    )
    parser.add_argument(
        "--cython",
        action="store_true",
        help="Turn on Cython for accelerated computation",
    )

    return parser


def run(filename="input.json", path=".", **kwargs):
    """Main function for running despasito calculations.

    All inputs and settings should be in the supplied JSON file(s).

    Parameters
    ----------
    filename : str, Optional, default="input.json"
        Input file containing instructions for various aspects of the calculation
    path : str, Optional, default="."
        Path to input file
    kwargs
        Keywords for other aspects of calculation
    """

    np.seterr(divide="ignore")
    # np.warnings.filterwarnings('error', category=np.VisibleDeprecationWarning)

    # read input file (need to add command line specification)
    logger.info("Begin processing input file: %s" % filename)
    eos_dict, thermo_dict, output_file = read_input.extract_calc_data(filename, path, **kwargs)

    thermo_dict["MultiprocessingObject"] = kwargs["MultiprocessingObject"]

    if output_file:
        file_dict = {"output_file": output_file}
    else:
        file_dict = {"output_file": "despasito_out.txt"}

    logger.debug("EOS dict:")
    for key, value in eos_dict.items():
        logger.debug("    {}: {}".format(key, value))
    logger.debug("Thermo dict:")
    for key, value in thermo_dict.items():
        logger.debug("    {}: {}".format(key, value))
    logger.info("Finish processing input file: {}".format(filename))

    # Run either parametrization or thermodynamic calculation
    fitting_opts = ["objective_method", "nan_number", "nan_ratio"]

    if "optimization_parameters" in thermo_dict:
        for key, exp_dict in thermo_dict["exp_data"].items():
            eos_dict = exp_dict["eos_dict"]
            thermo_dict["exp_data"][key].pop("eos_dict", None)
            thermo_dict["exp_data"][key]["eos_obj"] = initiate_eos(**eos_dict)
            for key2 in fitting_opts:
                if key2 in thermo_dict:
                    thermo_dict["exp_data"][key][key2] = thermo_dict[key2]
        logger.info("Initializing parametrization procedure")

        output_dict = fit(**thermo_dict.copy())
        output_dict.update(
            {
                "fit_bead": thermo_dict["optimization_parameters"]["fit_bead"],
                "fit_parameter_names": thermo_dict["optimization_parameters"]["fit_parameter_names"],
            }
        )
        logger.info("Finished parametrization")
        write_output.writeout_fit_dict(output_dict, **file_dict)
    else:
        Eos = initiate_eos(**eos_dict)
        logger.info("Initializing thermodynamic calculation")
        output_dict = thermo(Eos, **thermo_dict.copy())
        logger.info("Finished thermodynamic calculation")
        try:
            write_output.writeout_thermo_dict(output_dict, thermo_dict["calculation_type"], **file_dict)
        except Exception:
            logger.info("Final Output: {}".format(output_dict))

    if thermo_dict["MultiprocessingObject"].flag_use_mp:
        thermo_dict["MultiprocessingObject"].end_pool()



""" Routines for writing .txt and .json output files files from dictionaries.
"""

# import logging
import json
import collections
import numpy as np

# logger = logging.getLogger(__name__)


def write_EOSparameters(library, filename):
    """
    Sort and export dictionary of input parameters into file in JSON format.

    Parameters
    ----------
    library : dict
        Dictionary of parameters to be sorted and exported.
    filename : str
        Filename (with or without path) of parameter files in JSON format.

    """

    # sort and write SAFT dict
    for i in library:
        library[i] = collections.OrderedDict(sorted(list(library[i].items()), key=lambda tup: tup[0].lower()))
    f = open(filename, "w")
    json.dump(library, f, indent=4)


def writeout_thermo_dict(output_dict, calctype, output_file="thermo_output.txt"):
    """
    Write out result of thermodynamic calculation.

    This file is automatically saved for a thermodynamic calculation (not parameter
    fitting) using the default ``output_file`` name. Import dictionary of both input
    and output data to produce a file. A line in the top clarifies the calculation
    type done. Each system input and output properties (e.g. T, P, etc.) are expressed
    as a columns.

    Parameters
    ----------
    output_dict : dict
        Dictionary of given and calculated information from thermodynamic module
    calculation_type : str
        Thermodynamic calculation type used
    output_file : str, Optional, default="thermo_output.txt"
        Name of output file

    """

    # Define units
    units = {
        "T": "K",
        "P": "Pa",
        "Psat": "Pa",
        "rhol": "mol/m^3",
        "rhov": "mol/m^3",
        "delta": "Pa^(1/2)",
    }

    # Make comment line
    comment = "# This data was generated in DESPASITO using the thermodynamic "
    "calculation: {}".format(calctype)

    # Make results matrix
    keys = []
    matrix = []
    for key, value in output_dict.items():
        if np.size(value[0]) > 1:
            tmp_matrix = np.transpose(np.stack(value))
        else:
            if len(np.shape(value[0])) == 0:
                tmp_matrix = np.array(value)
            else:
                tmp_matrix = np.concatenate(value, axis=0)

        if len(tmp_matrix.shape) == 1:
            keys.append(key)
            matrix.append(np.array(tmp_matrix))
        else:
            for i in range(len(tmp_matrix)):
                keys.append(key + str(i + 1))
                matrix.append(np.array(tmp_matrix[i]))
    matrix = np.transpose(np.stack(matrix))

    # Make header line
    header = "#"
    for key in keys:
        if key in units:
            unit = " [{}]".format(units[key])
        else:
            unit = ""
        header += " {}{},".format(key, unit)

    # Write to file
    with open(output_file, "w") as f:
        f.write(comment + "\n")
        f.write(header + "\n")
        for row in matrix:
            f.write((" {}," * len(row)).format(*row) + "\n")


def writeout_fit_dict(output_dict, output_file="fit_output.txt"):
    """
    Write out result of fitting calculation.

    Parameters
    ----------
    output_dict : dict
        Dictionary of given and calculated information from thermodynamic module.
    output_file : str, Optional, default="fit_output.txt"
        Name of output file

    """

    header = (
        "DESPASITO was used to fit parameters for the bead {} Obj. Value:"
        " {}\n".format(output_dict["fit_bead"], output_dict["objective_value"]) + "Parameter, Value\n"
    )
    with open(output_file, "w") as f:
        f.write(header)
        for i in range(len(output_dict["fit_parameter_names"])):
            f.write(
                "{}, {}\n".format(
                    output_dict["fit_parameter_names"][i],
                    output_dict["parameters_final"][i],
                )
            )



"""
Unit and regression test for the despasito package.
"""

# Import package, test suite, and other packages as needed
import despasito.equations_of_state.cubic.peng_robinson

import pytest
import sys
import numpy as np

xi = np.array([0.827, 0.173])
molecular_composition = np.array([[1.0, 0.0], [0.0, 1.0]])
beads = ["acetone", "chloroform"]
bead_library = {
    "acetone": {"Tc": 508.1, "Pc": 4690000.0, "omega": 0.304},
    "chloroform": {"Tc": 536.4, "Pc": 5471550.0, "omega": 0.221902},
}
cross_library = {"acetone": {"chloroform": {"kij": -0.0605}}}

Eos = despasito.equations_of_state.initiate_eos(
    eos="cubic.peng_robinson",
    xi=xi,
    beads=beads,
    molecular_composition=molecular_composition,
    bead_library=bead_library,
    cross_library=cross_library,
)
T = 332.15
P = np.array([101330.0])
yi = np.array([0.89, 0.11])
rho = np.array([12546.22])


def test_peng_robinson_imported():
    #    """Sample test, will always pass so long as import statement worked"""
    assert "despasito.equations_of_state.cubic.peng_robinson" in sys.modules


def test_PR_coefficients(
    xi=xi,
    beads=beads,
    molecular_composition=molecular_composition,
    bead_library=bead_library,
):  # Test ability to create EOS object without association sites
    Eos_class = despasito.equations_of_state.initiate_eos(
        eos="cubic.peng_robinson",
        xi=xi,
        beads=beads,
        molecular_composition=molecular_composition,
        bead_library=bead_library,
    )
    tmp = [
        Eos_class.bead_library[beads[0]]["kappa"],
        Eos_class.bead_library[beads[1]]["kappa"],
    ]
    assert (
        Eos_class.eos_dict["ai"] == pytest.approx(np.array([1.73993846, 1.66217026]), abs=1e-4)
        and Eos_class.eos_dict["bi"] == pytest.approx(np.array([7.00758212e-05, 6.34118233e-05]), abs=1e-9)
        and tmp == pytest.approx(np.array([0.81854211, 0.70357958]), abs=1e-4)
    )


def test_peng_robinson_pressure(xi=xi, T=T, Eos=Eos, rho=rho):
    #   """Test ability to predict P with association sites"""
    P = Eos.pressure(rho, T, xi)[0]
    assert P == pytest.approx(69904905.698, abs=1e-1)


def test_peng_robinson_class_assoc_mu(P=P, xi=xi, T=T, Eos=Eos, rho=rho):
    #   """Test ability to predict P with association sites"""
    phi = Eos.fugacity_coefficient(P, rho, xi, T)
    #    assert mui == pytest.approx(np.array([1.61884825, -4.09022886]),abs=1e-4)
    assert phi == pytest.approx(np.array([1.12643785, 0.55712584]), abs=1e-4)



# cython: profile=True
import numpy as np
cimport numpy as np
from libc.stdio cimport printf

FTYPE = np.float64
ctypedef np.float_t FTYPE_t

cdef double const_molecule_per_nm3 = 6.02214086e-4 # mol/m^3 to molecules/nm^3 

def calc_Xika(indices, rho, xi, nui, nk, Fklab, Kklab, gr_assoc): # , maxiter=500, tol=1e-12, damp=.1
    r""" 
    Calculate the fraction of molecules of component i that are not bonded at a site of type a on group k.

    Parameters
    ----------
    indices : list[list]
        A list of sets of (component, bead, site) to identify the values of the Xika matrix that are being fit
    rho : numpy.ndarray
        Number density of system [molecules/m^3]
    xi : numpy.ndarray
        Mole fraction of each component, sum(xi) should equal 1.0
    nui : numpy.array
        :math:`\nu_{i,k}/k_B`, Array of number of components by number of bead types. Defines the number of each type of group in each component. 
    nk : numpy.ndarray
        For each bead the number of each type of site
    Fklab : numpy.ndarray
        The association strength between a site of type a on a group of type k of component i and a site of type b on a group of type l of component j., known as the Mayer f-function.
    Kklab : numpy.ndarray
        Bonding volume between each association site
    gr_assoc : numpy.ndarray
        Reference fluid pair correlation function used in calculating association sites, (len(rho) x Ncomp x Ncomp)

    Returns
    -------
    Xika : numpy.ndarray
        The fraction of molecules of component i that are not bonded at a site of type a on group k. Matrix (len(rho) x Ncomp x Nbeads x len(sitenames))
    err_array : numpy.ndarray
        Of the same length of rho, is a list in the error of the total error Xika for each point. 
    """
    Xika_init = 0.5*np.ones(len(indices), dtype=np.float_)

    indices = indices.astype(np.intc)
    rho = rho.astype(np.float_)
    xi = xi.astype(np.float_)
    nui = nui.astype(np.float_)
    nk = nk.astype(np.float_)
    Fklab = Fklab.astype(np.float_)
    Kklab = Kklab.astype(np.float_)
    gr_assoc = gr_assoc.astype(np.float_)

    l_K = len(np.shape(Kklab))
    if l_K == 4:
        Xika_final, err_array = calc_Xika_4(indices, rho, xi, nui, nk, Fklab, Kklab, gr_assoc, Xika_init)
    if l_K == 6:
        Xika_final, err_array = calc_Xika_6(indices, rho, xi, nui, nk, Fklab, Kklab, gr_assoc, Xika_init)

    return Xika_final, err_array

cdef calc_Xika_4(int[:,:] indices, double[:] rho, double[:] xi, double[:,:] nui, double[:,:] nk, double[:,:,:,:] Fklab, double[:,:,:,:] Kklab, double[:,:,:] gr_assoc, double[:] Xika_elements):

    cdef int  maxiter=500
    cdef double  tol=1e-12
    cdef double  damp=0.1

    cdef int nrho = rho.shape[0]
    cdef int l_ind = indices.shape[0]

    cdef int r, knd, ind, iind, i, k, a, jjnd, jnd, j, l, b, z
    cdef double delta, obj, Xika_max
    cdef double[:] Xika_elements_new = np.empty(l_ind, dtype=np.float_)

    cdef np.ndarray[np.float_t, ndim=2] Xika_final = np.empty((nrho, l_ind), dtype=np.float_)
    cdef np.ndarray[np.float_t, ndim=1] err_array = np.empty(nrho, dtype=np.float_)

    for r in range(nrho):
        for knd in range(maxiter):
            Xika_elements_new[:] = 1.0
            ind = 0
            for iind in range(l_ind):
                i = indices[iind][0]
                k = indices[iind][1]
                a = indices[iind][2]
                jnd = 0
                for jjnd in range(l_ind):
                    j = indices[jjnd][0] 
                    l = indices[jjnd][1]
                    b = indices[jjnd][2]
                    delta = Fklab[k, l, a, b] * Kklab[k, l, a, b] * gr_assoc[r,i, j]
                    Xika_elements_new[ind] = Xika_elements_new[ind] + (
                        const_molecule_per_nm3 
                        * rho[r] 
                        * xi[j] 
                        * nui[j,l] 
                        * nk[l,b] * 
                        Xika_elements[jnd] 
                        * delta
                    )
                    #printf("[%.14e %.14e %.14e %.14e %.14e %.14e]\n",rho[r], xi[j], nui[j,l], nk[l,b], Xika_elements[jnd], delta)
                    jnd += 1
                ind += 1

            obj = 0
            Xika_max = 0
            for z in range(l_ind):
                Xika_elements_new[z] = 1.0/Xika_elements_new[z]
                obj += abs(Xika_elements_new[z]-Xika_elements[z])
                if Xika_elements_new[z] > Xika_max:
                    Xika_max = Xika_elements_new[z]

            if obj < tol:
                break
            else:
                if obj/Xika_max > 1e+3:
                    for z in range(l_ind):
                        Xika_elements[z] = Xika_elements[z] + damp*(Xika_elements_new[z] - Xika_elements[z])
                else:
                    for z in range(l_ind):
                        Xika_elements[z] = Xika_elements_new[z]

        err_array[r] = obj
        Xika_final[r,:] = Xika_elements

    return Xika_final, err_array

cdef calc_Xika_6(int[:,:] indices, double[:] rho, double[:] xi, double[:,:] nui, double[:,:] nk, double[:,:,:,:] Fklab, double[:,:,:,:,:,:] Kklab, double[:,:,:] gr_assoc, double[:] Xika_elements):

    cdef int  maxiter=500
    cdef double  tol=1e-12
    cdef double  damp=0.1

    cdef int nrho = rho.shape[0]
    cdef int l_ind = indices.shape[0]

    cdef int r, knd, ind, iind, i, k, a, jnd, jjnd, j, l, b, z
    cdef double delta, obj, Xika_max
    cdef double[:] Xika_elements_new = np.empty(l_ind, dtype=np.float_)

    cdef np.ndarray[np.float_t, ndim=2] Xika_final = np.empty((nrho, l_ind), dtype=np.float_)
    cdef np.ndarray[np.float_t, ndim=1] err_array = np.empty(nrho, dtype=np.float_)

    for r in range(nrho):
        for knd in range(maxiter):
            Xika_elements_new[:] = 1.0
            ind = 0
            for iind in range(l_ind):
                i = indices[iind][0]
                k = indices[iind][1]
                a = indices[iind][2]
                jnd = 0
                for jjnd in range(l_ind):
                    j = indices[jjnd][0]
                    l = indices[jjnd][1]
                    b = indices[jjnd][2]
                    delta = Fklab[k, l, a, b] * Kklab[i, j, k, l, a, b] * gr_assoc[r,i, j]
                    Xika_elements_new[ind] = Xika_elements_new[ind] + (
                        const_molecule_per_nm3 
                        * rho[r] 
                        * xi[j] 
                        * nui[j,l] 
                        * nk[l,b] 
                        * Xika_elements[jnd] 
                        * delta
                    )
                    jnd += 1
                ind += 1

            obj = 0
            Xika_max = 0
            for z in range(l_ind):
                Xika_elements_new[z] = 1.0 / Xika_elements_new[z]
                obj += abs(Xika_elements_new[z]-Xika_elements[z])
                if Xika_elements_new[z] > Xika_max:
                    Xika_max = Xika_elements_new[z]

            if obj < tol:
                break
            else:
                if obj/Xika_max > 1e+3:
                    for z in range(l_ind):
                        Xika_elements[z] = Xika_elements[z] + damp*(Xika_elements_new[z] - Xika_elements[z])
                else:
                    for z in range(l_ind):
                        Xika_elements[z] = Xika_elements_new[z]

        err_array[r] = obj
        Xika_final[r,:] = Xika_elements

    return Xika_final, err_array




""" Routines for parsing input files in JSON format to dictionaries for program use.
"""

import logging
import json
import numpy as np
import os

logger = logging.getLogger(__name__)


def append_data_file_path(input_dict, path="."):
    r"""
    Appends path from command line option to data file(s).

    Parameters
    ----------
    input_dict
        Dictionary of input data
    path: str
        relative path to append to existing data files
    """
    if "EOSgroup" in input_dict:
        input_dict["EOSgroup"] = os.path.join(path, input_dict["EOSgroup"])

    if "EOScross" in input_dict:
        input_dict["EOScross"] = os.path.join(path, input_dict["EOScross"])

    for key, val in input_dict.items():
        if isinstance(val, dict) and "file" in val:
            input_dict[key]["file"] = os.path.join(path, input_dict[key]["file"])


def json_to_dict(filename):
    r"""
    Extract contents of JSON formatted file as a dictionary

    Parameters
    ----------
    filename : str
        File name and path leading to file (in JSON format) location

    Returns
    -------
    dictionary : dict
        Dictionary resulting from file
    """

    with open(filename, "r") as f:
        output = f.read()
    dictionary = json.loads(output)

    return dictionary


def extract_calc_data(input_fname, path=".", **thermo_dict):
    r"""
    Parse dictionary from input file in JSON format into a dictionary.

    Resulting dictionaries are used for creating the equation of state object, and for
    passing instructions for thermodynamic calculations.

    Parameters
    ----------
    input_fname : str
        The file name of a file in JSON format in the current directory containing
        (1) the paths to equation of state parameters,
        (2) :mod:`~despasito.thermodynamics.calculation_types` and inputs for
        thermodynamic calculations (e.g. density options for
        :func:`~despasito.thermodynamics.calc.pressure_vs_volume_arrays`).
    path : str, Optional, default="."
        Path to ``input_fname``
    thermo_dict
        Additional keyword arguments

    Returns
    -------
    eos_dict : dict
        Dictionary of keyword arguments representing bead definitions and parameters
        used to later initialize Eos object.
        :func:`despasito.equations_of_state.initiate_eos`
    thermo_dict : dict
        Dictionary of keyword arguments for thermodynamic calculations or parameter
        fitting. :func:`despasito.thermodynamics.thermo`
    output_file : str
        Output from calculation. Default is None, but an alternative can be defined
        as output_file keyword argument.
    """

    # Extract dictionary from input file
    input_dict = json_to_dict(input_fname)

    # Look for data (library) files in the user-supplied path
    append_data_file_path(input_dict, path)

    if "output_file" in input_dict:
        output_file = input_dict["output_file"]
    else:
        output_file = None

    # Make bead data dictionary for EOS
    # process input file
    if "bead_configuration" in input_dict:
        beads, molecular_composition = process_bead_data(input_dict["bead_configuration"])
        eos_dict = {"beads": beads, "molecular_composition": molecular_composition}
    elif "optimization_parameters" in input_dict:
        eos_dict = {}
    else:
        raise ValueError("Bead configuration line is missing for thermodynamic calculation.")

    # read EOS groups file
    eos_dict["bead_library"] = json_to_dict(input_dict["EOSgroup"])
    if "EOScross" in input_dict:
        eos_dict["cross_library"] = json_to_dict(input_dict["EOScross"])
        logger.info("Cross interaction parameters have been accepted")
    else:
        logger.info("No EOScross file specified")

    # Extract relevant system state inputs
    EOS_dict_keys = ["bead_configuration", "EOSgroup", "EOScross", "output_file"]
    for key, value in input_dict.items():
        if key.startswith("eos_"):
            new_key = "_".join(key.split("_")[1:])
            eos_dict[new_key] = input_dict[key]
        elif key == "eos":
            eos_dict["eos"] = input_dict["eos"]
        elif key not in EOS_dict_keys:
            thermo_dict[key] = value

    for key in ["numba", "cython", "python"]:
        if key in thermo_dict:
            eos_dict[key] = thermo_dict[key]
            del thermo_dict[key]

    if "optimization_parameters" not in thermo_dict:
        logger.info(
            "The following thermo calculation parameters have been "
            "provided: {}\n".format((", ".join(thermo_dict.keys())))
        )
    else:  # parameter fitting
        thermo_dict = process_param_fit_inputs(thermo_dict)
        for exp_key in thermo_dict["exp_data"]:
            if "eos_dict" not in thermo_dict["exp_data"][exp_key]:
                thermo_dict["exp_data"][exp_key]["eos_dict"] = {}
            for key, value in eos_dict.items():
                if key not in thermo_dict["exp_data"][exp_key]["eos_dict"]:
                    thermo_dict["exp_data"][exp_key]["eos_dict"][key] = value
        tmp = ""
        for key, value in thermo_dict["exp_data"].items():
            tmp += " {} ({}),".format(key, value["data_class_type"])
        logger.info(
            "The bead, {}, will have the parameters {}, fit using the following "
            "data:\n {}".format(
                thermo_dict["optimization_parameters"]["fit_bead"],
                thermo_dict["optimization_parameters"]["fit_parameter_names"],
                tmp,
            )
        )

    return eos_dict, thermo_dict, output_file


def file2paramdict(filename, delimiter=" "):
    r"""
    Converted text file into a dictionary.

    Each line in the input file is a key followed by a value in the resulting
    dictionary.

    Parameters
    ----------
    filename : str
        File of keys and values
    delimiter : str, Optional, default=" "
        String separating key and value within file

    Returns
    -------
    dictionary : dict
        Resulting dictionary
    """

    dictionary = {}
    with open(filename, "r") as filedata:
        for line in filedata:
            line.rstrip()
            linearray = line.split(delimiter)
            if len(linearray) == 2:
                try:
                    dictionary[linearray[0]] = eval(linearray[1])
                except Exception:
                    if line[1].isdigit():
                        dictionary[linearray[0]] = float(linearray[1])
                    else:
                        dictionary[linearray[0]] = linearray[1]

    return dictionary


def process_bead_data(bead_data):
    r"""
    Process system information from input file.

    Parameters
    ----------
    bead_data : list[list]
        List of molecules in the system. Each molecule is represented as a list of
        beads and each bead is represented as a list with the bead name followed by an
        integer with the number of beads in that molecule.

    Returns
    -------
    beads : list[str]
        List of unique bead names used among components
    molecular_composition : numpy.ndarray
        Array of number of components by number of bead types. Defines the number of
        each type of group in each component.
    """

    # find list of unique beads
    beads = []
    for i in range(len(bead_data)):
        for j in range(len(bead_data[i])):
            if bead_data[i][j][0] not in beads:
                beads.append(bead_data[i][j][0])
    beads.sort()

    molecular_composition = np.zeros((len(bead_data), len(beads)))
    for i in range(len(bead_data)):
        for j in range(len(bead_data[i])):
            for k in range(np.size(beads)):
                if bead_data[i][j][0] == beads[k]:
                    molecular_composition[i, k] = bead_data[i][j][1]
    return beads, molecular_composition


def process_param_fit_inputs(thermo_dict):
    r"""
    Process parameter fitting information.

    Parameters
    ----------
    thermo_dict : dict
        Dictionary of instructions for thermodynamic calculations or parameter
        fitting. This dictionary is directly from the input file.

        - optimization_parameters (dict) - Parameters used in basin fitting algorithm

            - fit_bead (str) - Name of bead whose parameters are being fit, should be
            in bead list of bead_configuration
            - fit_parameter_names (list[str]) - This list of contains the name of the
            parameter being fit (e.g. epsilon). See EOS documentation for limitations
            on supported parameter names. Cross interaction parameter names should be
            composed of parameter name and the other bead type, separated by an
            underscore (e.g. epsilon_CO2).
            - \*_bounds (list[float]), Optional - This list contains the minimum and
            maximum of a parameter from fit_parameter_names, represented by the
            asterisk. See input file instructions for more information.
            - parameters_guess (list[float]), Optional - Initial guess in parameters
            being fit. Should be the same length at fit_parameter_names and contain a
            reasonable guess for each parameter. If this is not provided, a guess is
            made based on the type of parameter from the Eos object.

        - *name* (dict) - Dictionary representing a data set that the parameters are
        fit to. Each dictionary is added to the exp_data dictionary before being passed
        to the fitting algorithm. Each *name* is used as the key in exp_data. *name* is
        an arbitrary string used to identify the data set and used later in reporting
        objective function values during the fitting process. See data type objects for
        more details.

            - file (str) - File of experimental data
            - data_class_type (str) - One of the supported data type objects to fit
            parameters
            - calculation_type (str) - One of the supported thermo calculation types
            that would be associated with the chosen data_class_type

    Returns
    -------
    new_thermo_dict : dict
        Dictionary of instructions for thermodynamic calculations or parameter fitting.
        This dictionary is reformatted and includes imported data. Dictionary values
        below are altered before being passed on, all other key and value sets are
        blindly passed.

        - optimization_parameters (dict) - Parameters used in basin fitting algorithm

            - fit_bead (str) - Name of bead whose parameters are being fit, should be
            in bead list of bead_configuration
            - fit_parameter_names (list[str]) - This list of contains the name of the
            parameter being fit (e.g. epsilon). See EOS documentation for restrictions
            on supported parameter names. Cross interaction parameter names should be
            composed of parameter name and the other bead type, separated by an
            underscore (e.g. epsilon_CO2).
            - parameters_guess (list[float]), Optional - Initial guess in parameter.
            If one is not provided, a guess is made based on the type of parameter
            from Eos object.
            - \*_bounds (list[float]), Optional - This list contains the minimum and
            maximum of a parameter in fit_parameter_names, represented in place of the
            asterisk. See input file instructions for more information.

        - exp_data (dict) - This dictionary is made up of a dictionary for each data
        set that the parameters are fit to. Each key is an arbitrary string used to
        identify the data set and used later in reporting objective function values
        during the fitting process. See data type objects for more details.

            - data_class_type (obj) - One of the supported data type objects to fit
            parameters
    """

    # Initial new dictionary that will have dictionary for extracted data
    new_thermo_dict = {"exp_data": {}}

    for key, value in thermo_dict.items():
        if isinstance(value, dict) and "data_class_type" in value:
            new_thermo_dict["exp_data"][key] = process_exp_data(value)
        else:
            new_thermo_dict[key] = value

    test1 = set(["exp_data", "optimization_parameters"]).issubset(list(new_thermo_dict.keys()))
    test2 = set(["fit_bead", "fit_parameter_names"]).issubset(list(new_thermo_dict["optimization_parameters"].keys()))
    if not all([test1, test2]):
        raise ValueError(
            "An exp_data dictionary (dictionary with 'data_class_type' key) as well as"
            " an optimization_parameters dictionary with 'fit_bead' and "
            "'fit_parameter_names' must be provided."
        )

    return new_thermo_dict


def process_exp_data(exp_data_dict):
    r"""
    Convert experimental data dictionary into form used by parameter fitting module.

    Note that there should be one dictionary per data set.

    Parameters
    ----------
    exp_data : dict
       This dictionary is made up of a dictionary for each data set that the parameters
       are fit to. Each key is an arbitrary string used to identify the data set and
       used later in reporting objective function values during the fitting process.
       See data type objects for more details. Dictionary values below are altered
       before being passed on, all other key and value sets are blindly passed on.

       - data_class_type (str) - One of the supported data type objects to fit
       parameters
       - file (str) - File name in current working directory, or path to desired
       experimental data. See experimental data page for examples of acceptable format.

    Returns
    -------
    exp_data : dict
        Reformatted dictionary of experimental data. This dictionary is made up of
        a dictionary for each data set that the parameters are fit to. Each key is
        an arbitrary string used to identify the data set and used later in reporting
        objective function values during the fitting process. See data type objects
        for more details. Dictionary values below are altered from input dictionary,
        all other key and value sets are blindly passed on, or extracted from data
        file with process_exp_data_file function.

        - data_class_type (str) - One of the supported data type objects to fit
        parameters
    """

    exp_data = {}
    for key, value in exp_data_dict.items():
        if key == "data_class_type":
            exp_data["data_class_type"] = value
        elif key == "file":
            file_dict = process_exp_data_file(value)
            exp_data.update(file_dict)
        elif key == "bead_configuration":
            beads, molecular_composition = process_bead_data(value)
            exp_data["eos_dict"] = {
                "beads": beads,
                "molecular_composition": molecular_composition,
            }
        else:
            exp_data[key] = value

    return exp_data


def process_exp_data_file(fname):
    r"""
    Import data file and convert columns into dictionary entries.

    The headers in the file are the dictionary keys. The top line is skipped, and
    column headers are the second line. Note that column headers should be thermo
    properties defined in :ref:`data-types` (e.g. T, P, x1, x2, y1, y2). Mole
    fractions x1, x2, ... should be in the same order as in the bead_configuration
    line of the input file. No mole fractions should be left out.

    Parameters
    ----------
    fname : str
        File name or path to experimental data file

    Returns
    -------
    file_dict : dict
        Dictionary of experimental data from file.
    """

    try:
        data = np.transpose(np.genfromtxt(fname, delimiter=",", names=True, skip_header=1))
    except Exception:
        raise ValueError("Cannot import '{}', Check data file formatting.".format(fname))
    file_dict = {name: data[name] for name in data.dtype.names}

    # Sort through properties
    key_del = []
    xi, yi, zi = [[], [], []]
    for key, value in file_dict.items():
        if "#" in key:
            key.replace("#", "").replace(" ", "")

        # Assuming mole fractions are listed starting at x1 and continue in order
        if key.startswith("x"):
            xi.append(value)
        elif key.startswith("y"):
            yi.append(value)
        elif key.startswith("z"):
            zi.append(value)
        else:
            continue
        key_del.append(key)

    for key in key_del:
        file_dict.pop(key, None)

    if xi:
        file_dict["xi"] = np.transpose(np.array([np.array(x) for x in xi]))

    if yi:
        file_dict["yi"] = np.transpose(np.array([np.array(y) for y in yi]))

    if zi:
        file_dict["zi"] = np.transpose(np.array([np.array(z) for z in zi]))

    return file_dict



"""
Contains constants used in the other files.
"""

# physical constants

kb = 1.38064852e-23  # Boltzmann constant J/K
me = 9.10938356e-31  # mass of electron kg
h = 6.62607004e-34  # Planck's constant J*S
Nav = 6.02214086e23  # Avogadro's number
R = 8.31446261815324  # Gas constant J/ ( mol*K)
molecule_per_nm3 = 6.02214086e-4  # mol/m^3 to molecules/nm^3
m2nm = 1e9  # meters to nm
Atometer = 1.0e-10  # conversion of angstroms to meters



# -- coding: utf8 --

r"""
EOS object for SAFT ideal gas contributions to the Helmholtz energy with
:func:`~despasito.equations_of_state.saft.Aideal.Aideal_contribution`
"""

import numpy as np
import logging

from despasito.equations_of_state import constants
import despasito.equations_of_state.eos_toolbox as tb

logger = logging.getLogger(__name__)


def Aideal_contribution(rho, T, xi, massi, method="Abroglie"):
    r"""
    Return a vector of ideal contribution of the Helmholtz energy.

    :math:`\frac{A^{ideal}}{N k_{B} T}`

    Supported methods include:
    :func:`~despasito.equations_of_state.saft.Aideal.Abroglie`,

    Parameters
    ----------
    rho : numpy.ndarray
        Number density of system [:math:`mol/m^3`]
    T : float
        Temperature of the system [K]
    xi : numpy.ndarray
        Mole fraction of each component, sum(xi) should equal 1.0
    massi : numpy.ndarray
        Vector of component masses that correspond to the mole fractions in xi [kg/mol]
    method : str, Optional, default=Abroglie
        The function name of the method to calculate the ideal contribution of the
        Helmholtz energy. To add a new one, add a function to:
        despasito.equations_of_state.saft.Aideal.py

    Returns
    -------
    Aideal : numpy.ndarray
        Helmholtz energy of ideal gas for each density given.
    """

    functions = {"Abroglie": Abroglie}

    if method in functions:
        function = functions[method]
    else:
        raise ValueError("Method, {}, was not found to calculate Aideal.".format(method))

    return function(rho, T, xi, massi)


def Abroglie(rho, T, xi, massi):
    r"""
    Return a vector of ideal contribution of Helmholtz energy derived from Broglie
    wavelength

    Parameters
    ----------
    rho : numpy.ndarray
        Number density of system [:math:`mol/m^3`]
    T : float
        Temperature of the system [K]
    xi : numpy.ndarray
        Mole fraction of each component, sum(xi) should equal 1.0
    massi : numpy.ndarray
        Vector of component masses that correspond to the mole fractions in xi [kg/mol]

    Returns
    -------
    Aideal : numpy.ndarray
        Helmholtz energy of ideal gas for each density given.
    """

    rho2 = rho * constants.molecule_per_nm3

    xi_tmp, massi_tmp = tb.remove_insignificant_components(xi, massi)

    # rhoi: (number of components,number of densities) number density of each component
    # for each density

    rhoi = np.outer(rho2, xi_tmp)
    Lambda = np.sqrt(
        (constants.h * constants.Nav * constants.m2nm)
        * (constants.h / constants.kb * constants.m2nm)
        / (2.0 * np.pi * massi_tmp * T)
    )
    log_broglie3_rho = np.log(Lambda**3 * rhoi)

    if np.isnan(np.sum(np.sum(xi_tmp * log_broglie3_rho, axis=1))):
        raise ValueError(
            "Aideal has values of zero when taking the log. All mole fraction values "
            "should be nonzero. Mole fraction: {}".format(xi_tmp)
        )
    else:
        Aideal = np.sum(xi_tmp * log_broglie3_rho, axis=1) - 1.0

    return Aideal



""" Optional constraint functions extracted for
:func:`~despasito.parameter_fitting.fit_functions.global_minimization` method.
"""


def type1(x, *args):
    """Example constraint. Notice that if this is used by a constraint class, there
    are no other args, while there are for dict constraints."""

    output = None

    return output



"""

Parallelization class to handle processing threads and logging.

"""

import numpy as np
import multiprocessing
import logging
import logging.handlers
import os
import glob

logger = logging.getLogger(__name__)


class MultiprocessingJob:
    """
    This object initiates the pool for multiprocessing jobs.

    Parameters
    ----------
    ncores : int, Optional, default=-1
        Number of processes used. If the default value of -1, the system cpu count
        is used.

    Attributes
    ----------
    pool : function
        This pool is used to parallelize jobs.
    """

    def __init__(self, ncores=-1):

        multiprocessing.freeze_support()
        self.flag_use_mp = True
        if ncores == -1:
            ncores = multiprocessing.cpu_count()  # includes logical cores!
            logger.info("Detected {} cores".format(ncores))
        elif ncores > 1:
            logger.info("Number of cores set to {}".format(ncores))
        elif ncores == 1:
            self.flag_use_mp = False
            logger.info("Number of cores set to 1, bypassing mp and using serial methods")
        else:
            raise ValueError("Number of cores cannot be zero or negative.")

        self.ncores = ncores

        if self.flag_use_mp:

            # Remove old mp logs
            self._extract_root_logging()

            # Initiate multiprocessing
            ctx = multiprocessing.get_context("spawn")
            self._pool = ctx.Pool(
                ncores,
                initializer=self._initialize_mp_handler,
                initargs=(self._level, self._logformat),
            )

            self.logfiles = []
            for worker in self._pool._pool:
                filename = "mp-handler-{0}.log".format(worker.pid)
                self.logfiles.append(filename)
            logger.info("MP log files: {}".format(", ".join(self.logfiles)))

    def _extract_root_logging(self):
        """Swap root handlers defined in despasito.__main__ with process specific log
        handlers"""
        for handler in logging.root.handlers:
            if "baseFilename" in handler.__dict__:
                self._logformat = handler.formatter._fmt
                self._level = handler.level

        if not hasattr(self, "_logformat"):
            self._logformat = None
            self._level = None

    @staticmethod
    def _initialize_mp_handler(level, logformat):
        """Wraps the handlers in the given Logger with an MultiProcessingHandler.

        Parameters
        ----------
        level : int
            The verbosity level of logging information can be set to any supported
            representation of the
            `logging level <LOGGING>`_
            .
        logformat : str
            Formating of logging information can be set to any supported representation
            of the
            `formatting class <FORMAT>`_.

        .. _LOGGING: https://docs.python.org/3/library/logging.html#logging-levels/
        .. _FORMAT: https://docs.python.org/3/library/logging.html#logging.Formatter/
        """

        logger = logging.getLogger("despasito")

        pid = os.getpid()
        filename = "mp-handler-{0}.log".format(pid)
        handler = logging.handlers.RotatingFileHandler(filename)
        if level is not None:
            logger.setLevel(level)
            handler.setLevel(level)
        if logformat is not None:
            handler.setFormatter(logging.Formatter(logformat))

        logger.addHandler(handler)

    def pool_job(self, func, inputs):
        """
        This function will setup and dispatch thermodynamic or parameter fitting jobs.

        Parameters
        ----------
        func : function
            Function used in job
        inputs : list[tuple]
            Each entry of this list contains the input arguments for each job

        Returns
        -------
        output : tuple
            This structure contains the outputs of the jobs given

        """

        if self.flag_use_mp:
            output = zip(*self._pool.map(func, inputs))
            self._consolidate_mp_logs()
        else:
            logger.info("Performing task serially")
            output = self.serial_job(func, inputs)

        return output

    @staticmethod
    def serial_job(func, inputs):
        """
        This function will serially perform thermodynamic jobs.

        Parameters
        ----------
        func : function
            Function used in job
        inputs : tuple
            The input arguments for this job

        Returns
        -------
        output : tuple
            This structure contains the outputs of the jobs given

        """

        output = []
        for i, finput in enumerate(inputs):
            foutput = func(finput)
            output.append(foutput)
        output = np.array(output, dtype=object)

        return np.transpose(output)

    def _consolidate_mp_logs(self):
        """Consolidate multiprocessing logs into main log"""
        for i, fn in enumerate(self.logfiles):
            with open(fn) as f:
                logger.info("Log from thread {0}:\n{1}".format(i, f.read()))
            open(fn, "w").write("")

    def _remove_mp_logs(self):
        """Ensure all previous mp logs are removed"""
        for i, fn in enumerate(self.logfiles):
            os.remove(fn)

    def end_pool(self):
        """Close multiprocessing pool"""
        if self.flag_use_mp:
            self._pool.close()
            self._pool.join()
            self._remove_mp_logs()


def initialize_mp_handler(level, logformat):
    """Wraps the handlers in the given Logger with an MultiProcessingHandler.

    Parameters
    ----------
    level : int
        The verbosity level of logging information can be set to any supported
        representation of the
        `logging level <LOGGING>`_
        .
    logformat : str
        Formating of logging information can be set to any supported representation of
        the
        `formatting class <FORMAT>`_
        .
    """

    logger = logging.getLogger("despasito")

    pid = os.getpid()
    filename = "mp-handler-{0}.log".format(pid)
    handler = logging.handlers.RotatingFileHandler(filename)

    handler.setFormatter(logging.Formatter(logformat))
    handler.setLevel(level)

    logger.addHandler(handler)


def batch_jobs(func, inputs, ncores=1, logger=None):
    """
    This function will setup and dispatch thermodynamic jobs.

    Parameters
    ----------
    func : function
        Function used in job
    inputs : list
        Each entry of this list contains the input arguments for each job
    ncores : int, Optional, default=1
        Number of processes used.
    logger : class, Optional, default=None
        The logger object used.

    Returns
    -------
    output : tuple
        This structure contains the outputs of the jobs given

    """

    if logger is None:
        logger = logging.getLogger()

    root_handlers = logging.root.handlers
    for handler in root_handlers:
        if "baseFilename" in handler.__dict__:
            logformat = handler.formatter._fmt
            level = handler.level
    logging.root.handlers = []

    pool = multiprocessing.Pool(ncores, initializer=initialize_mp_handler, initargs=(level, logformat))

    output = zip(*pool.map(func, inputs))

    logging.root.handlers = root_handlers

    for i, fn in enumerate(glob.glob("./mp-handler-*.log")):
        with open(fn) as f:
            logger.info("Log from thread {0}:\n{1}".format(i, f.read()))
        os.remove(fn)

    return output
